import sys

import click

from exasol_integration_test_docker_environment.cli.cli import cli
from exasol_integration_test_docker_environment.cli.options.test_environment_options import (
    DEFAULT_DISK_SIZE,
    DEFAULT_MEM_SIZE,
    LATEST_DB_VERSION,
)


@cli.command()
@click.option(
    "--show-default-db-version",
    is_flag=True,
    help="Displays the defualt Docker DB Image version",
)
@click.option(
    "--show-default-mem-size",
    is_flag=True,
    help="Displays the defualt DB mem size",
)
@click.option(
    "--show-default-disk-size",
    is_flag=True,
    help="Displays the defualt DB disk size",
)
def environment(
    show_default_db_version: bool,
    show_default_mem_size: bool,
    show_default_disk_size: bool,
):
    """
    Displays the default configurations of the DB.
    """
    if show_default_db_version:
        print(f"Default Docker DB Image Version: {LATEST_DB_VERSION}")
    elif show_default_mem_size:
        print(f"Default DB mem size: {DEFAULT_MEM_SIZE}")
    elif show_default_disk_size:
        print(f"Default DB disk size: {DEFAULT_DISK_SIZE}")



from exasol_integration_test_docker_environment.lib.base.base_task import BaseTask
from exasol_integration_test_docker_environment.lib.docker.images.push.docker_image_push_task import (
    DockerPushImageTask,
)
from exasol_integration_test_docker_environment.lib.docker.images.task_creator_from_build_tasks import (
    TaskCreatorFromBuildTasks,
)


class PushTaskCreatorFromBuildTasks(TaskCreatorFromBuildTasks):

    def __init__(self, task: BaseTask) -> None:
        self.task = task

    def create_task_with_required_tasks(self, build_task, required_task_info):
        push_task = self.task.create_child_task_with_common_params(
            DockerPushImageTask,
            image_name=build_task.image_name,
            required_task_info=required_task_info,
        )
        return push_task



from enum import (
    Enum,
    auto,
)


class TaskState(Enum):
    NONE = auto()
    INIT = auto()
    AFTER_INIT = auto()
    RUN = auto()
    FINISHED = auto()
    CLEANUP = auto()
    CLEANED = auto()
    ERROR = auto()



from exasol_integration_test_docker_environment.lib.logging.abstract_log_handler import (
    AbstractLogHandler,
)
from exasol_integration_test_docker_environment.lib.models.config.log_config import (
    WriteLogFilesToConsole,
)


class ContainerLogHandler(AbstractLogHandler):

    def __init__(self, log_file_path, logger, description: str) -> None:
        super().__init__(log_file_path, logger)
        self.db_container_name = description

    def handle_log_line(self, log_line, error: bool = False) -> None:
        self._log_file.write(log_line + "\n")
        self._complete_log.append(log_line)

    def get_complete_log(self) -> list[str]:
        return self._complete_log

    def finish(self) -> None:
        if self._log_config.write_log_files_to_console == WriteLogFilesToConsole.all:
            self._logger.info(
                "Log %s\n%s", self.db_container_name, "\n".join(self._complete_log)
            )



from collections.abc import Generator
from pathlib import Path
from typing import (
    Union,
)

import luigi
from luigi import Task

from exasol_integration_test_docker_environment.lib.base.stoppable_base_task import (
    StoppableBaseTask,
)
from exasol_integration_test_docker_environment.lib.base.task_dependency import (
    DependencyState,
    DependencyType,
    TaskDependency,
    TaskDescription,
)


class DependencyLoggerBaseTask(StoppableBaseTask):

    def _get_dependencies_path_for_job(self) -> Path:
        return Path(super()._get_output_path_for_job(), "dependencies")

    def _get_dependencies_path(self) -> Path:
        path = Path(self._get_dependencies_path_for_job(), self.task_id)
        path.mkdir(parents=True, exist_ok=True)
        return path

    def _get_dependencies_requires_path(self) -> Path:
        return Path(self._get_dependencies_path(), "requires")

    def _get_dependencies_dynamic_path(self) -> Path:
        return Path(self._get_dependencies_path(), "dynamic")

    def handle_requires_value(
        self, tasks: Union[Task, list, tuple, dict]
    ) -> Union[Task, list, tuple, dict]:
        dependency_path = self._get_dependencies_requires_path()
        if not dependency_path.exists():
            with dependency_path.open("w") as dependencies_file:
                self.write_dependency(
                    dependencies_file=dependencies_file,
                    dependency_type=DependencyType.requires,
                    dependency_state=DependencyState.requested,
                    index=0,
                    value=tasks,
                )
                return tasks
        else:
            return tasks

    def handle_requires_generator(self, tasks):
        dependency_path = self._get_dependencies_requires_path()
        if not dependency_path.exists():
            with dependency_path.open("w") as dependencies_file:
                result = list(
                    self.write_dependencies_for_generator(
                        dependencies_file=dependencies_file,
                        task_generator=tasks,
                        dependency_type=DependencyType.requires,
                    )
                )
                return result
        else:
            return tasks

    def write_dependencies_for_generator(
        self, dependencies_file, task_generator, dependency_type: DependencyType
    ):
        index = 0
        try:
            element = next(task_generator)
            while True:
                self.write_dependency(
                    dependencies_file=dependencies_file,
                    dependency_type=dependency_type,
                    dependency_state=DependencyState.requested,
                    index=index,
                    value=element,
                )
                result = yield element
                element = task_generator.send(result)
                index += 1
        except StopIteration:
            pass

    def write_dependency(
        self,
        dependencies_file,
        dependency_type: DependencyType,
        dependency_state: DependencyState,
        index: int,
        value,
    ):
        for task in self.flatten_tasks(value):
            dependency = TaskDependency(
                source=self.get_task_description(),
                target=task.get_task_description(),
                type=dependency_type,
                index=index,
                state=dependency_state,
            )
            dependencies_file.write(f"{dependency.to_json()}")
            dependencies_file.write("\n")

    def requires(self):
        tasks = super().requires()
        if tasks is not None:
            if isinstance(tasks, (Task, list, tuple, dict)):
                return self.handle_requires_value(tasks)
            else:
                return self.handle_requires_generator(tasks)
        else:
            return []

    def run(self):
        dependency_path = self._get_dependencies_dynamic_path()
        with dependency_path.open("a") as dependencies_file:
            task_generator = super().run()
            if task_generator is not None:
                yield from self.write_dependencies_for_generator(
                    dependencies_file=dependencies_file,
                    task_generator=task_generator,
                    dependency_type=DependencyType.dynamic,
                )

    def get_task_description(self) -> TaskDescription:
        return TaskDescription(id=self.task_id, representation=str(self))

    def flatten_tasks(self, generator: Generator) -> list["DependencyLoggerBaseTask"]:
        return [
            task
            for task in luigi.task.flatten(generator)
            if isinstance(task, DependencyLoggerBaseTask)
        ]



import ssl
from time import sleep
from urllib.parse import quote_plus
from xmlrpc.client import ServerProxy

import luigi

from exasol_integration_test_docker_environment.lib.base.dependency_logger_base_task import (
    DependencyLoggerBaseTask,
)
from exasol_integration_test_docker_environment.lib.base.json_pickle_parameter import (
    JsonPickleParameter,
)
from exasol_integration_test_docker_environment.lib.models.data.database_credentials import (
    DatabaseCredentialsParameter,
)
from exasol_integration_test_docker_environment.lib.models.data.database_info import (
    DatabaseInfo,
)
from exasol_integration_test_docker_environment.lib.models.data.docker_network_info import (
    DockerNetworkInfo,
)
from exasol_integration_test_docker_environment.lib.test_environment.parameter.external_test_environment_parameter import (
    ExternalDatabaseHostParameter,
    ExternalDatabaseXMLRPCParameter,
)
from exasol_integration_test_docker_environment.lib.test_environment.ports import (
    Ports,
)


class SetupExternalDatabaseHost(
    DependencyLoggerBaseTask,
    ExternalDatabaseXMLRPCParameter,
    ExternalDatabaseHostParameter,
    DatabaseCredentialsParameter,
):
    environment_name: str = luigi.Parameter()
    network_info: DockerNetworkInfo = JsonPickleParameter(DockerNetworkInfo, significant=False)  # type: ignore
    attempt: int = luigi.IntParameter(default=1)

    def run_task(self) -> None:
        database_host = self.external_exasol_db_host
        if (
            self.external_exasol_db_host == "localhost"
            or self.external_exasol_db_host == "127.0.0.1"
        ):
            database_host = self.network_info.gateway
        self.setup_database()
        ports = Ports(
            database=self.external_exasol_db_port,
            bucketfs=self.external_exasol_bucketfs_http_port,
            ssh=self.external_exasol_ssh_port,
            bucketfs_https=self.external_exasol_bucketfs_https_port,
        )
        assert database_host is not None
        database_info = DatabaseInfo(host=database_host, ports=ports, reused=False)
        self.return_object(database_info)

    def setup_database(self) -> None:
        if self.external_exasol_xmlrpc_host is not None:
            # TODO add option to use unverified ssl
            cluster = self.get_xml_rpc_object()
            self.start_database(cluster)
            cluster.bfsdefault.editBucketFS(
                {
                    "http_port": int(self.external_exasol_bucketfs_http_port),
                    "https_port": int(self.external_exasol_bucketfs_https_port),
                }
            )
            try:
                cluster.bfsdefault.addBucket(
                    {
                        "bucket_name": "myudfs",
                        "public_bucket": True,
                        "read_password": self.bucketfs_write_password,
                        "write_password": self.bucketfs_write_password,
                    }
                )
            except Exception as e:
                self.logger.info(e)
            try:
                cluster.bfsdefault.addBucket(
                    {
                        "bucket_name": "jdbc_adapter",
                        "public_bucket": True,
                        "read_password": self.bucketfs_write_password,
                        "write_password": self.bucketfs_write_password,
                    }
                )
            except Exception as e:
                self.logger.info(e)

    def get_xml_rpc_object(self, object_name: str = "") -> ServerProxy:
        assert self.external_exasol_xmlrpc_user and self.external_exasol_xmlrpc_password
        uri = "https://{user}:{password}@{host}:{port}/{cluster_name}/{object_name}".format(
            user=quote_plus(self.external_exasol_xmlrpc_user),
            password=quote_plus(self.external_exasol_xmlrpc_password),
            host=self.external_exasol_xmlrpc_host,
            port=self.external_exasol_xmlrpc_port,
            cluster_name=self.external_exasol_xmlrpc_cluster_name,
            object_name=object_name,
        )
        server = ServerProxy(uri, context=ssl._create_unverified_context())
        return server

    def start_database(self, cluster: ServerProxy) -> None:
        storage = self.get_xml_rpc_object("storage")
        # wait until all nodes are online
        self.logger.info("Waiting until all nodes are online")
        all_nodes_online = False
        while not all_nodes_online:
            all_nodes_online = True
            for nodeName in cluster.getNodeList():  # type: ignore
                node_state: Dict[str, str] = self.get_xml_rpc_object(nodeName).getNodeState()  # type: ignore
                if node_state["status"] != "Running":
                    all_nodes_online = False
                    break
            sleep(5)
        self.logger.info("All nodes are online now")

        # start EXAStorage
        if not storage.serviceIsOnline():
            if storage.startEXAStorage() != "OK":
                self.logger.info("EXAStorage has been started successfully")
            else:
                raise Exception("Not able startup EXAStorage!\n")
        elif storage.serviceIsOnline():
            self.logger.info("EXAStorage already online; continuing startup process")

        # triggering database startup
        for databaseName in cluster.getDatabaseList():  # type: ignore
            database = self.get_xml_rpc_object("/db_" + quote_plus(str(databaseName)))
            if not database.runningDatabase():
                self.logger.info("Starting database instance %s" % str(databaseName))
                database.startDatabase()
            else:
                self.logger.info(
                    "Database instance %s already running" % str(databaseName)
                )



import copy

from exasol_integration_test_docker_environment.abstract_method_exception import (
    AbstractMethodException,
)
from exasol_integration_test_docker_environment.lib.base.dependency_logger_base_task import (
    DependencyLoggerBaseTask,
)
from exasol_integration_test_docker_environment.lib.docker.images.create.docker_image_analyze_task import (
    DockerAnalyzeImageTask,
)
from exasol_integration_test_docker_environment.lib.docker.images.create.docker_image_create_task import (
    DockerCreateImageTask,
    DockerCreateImageTaskWithDeps,
)
from exasol_integration_test_docker_environment.lib.docker.images.image_info import (
    ImageInfo,
    ImageState,
)
from exasol_integration_test_docker_environment.lib.docker.images.required_task_info import (
    RequiredTaskInfo,
    RequiredTaskInfoDict,
)
from exasol_integration_test_docker_environment.lib.models.config.build_config import (
    build_config,
)


class DockerBuildBase(DependencyLoggerBaseTask):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def _get_build_steps_to_rebuild(self):
        return set(build_config().force_rebuild_from)

    def get_goals(self) -> set[str]:
        raise AbstractMethodException()

    def get_default_goals(self) -> set[str]:
        raise AbstractMethodException()

    def get_goal_class_map(self) -> dict[str, DockerAnalyzeImageTask]:
        raise AbstractMethodException()

    def register_required(self):
        self.analyze_tasks_futures = self.register_dependencies(
            self.create_analyze_tasks()
        )

    def create_analyze_tasks(self) -> dict[str, DockerAnalyzeImageTask]:
        goals = set(self.get_goals())
        self.goal_class_map = self.get_goal_class_map()
        self.available_goals = set(self.goal_class_map.keys())
        self._check_if_build_steps_to_rebuild_are_valid_goals()
        final_goals = self.get_default_goals()
        if len(goals) != 0:
            final_goals = set(goals)
        if final_goals.issubset(self.available_goals):
            tasks = {goal: self.goal_class_map[goal] for goal in final_goals}
            return tasks
        else:
            difference = goals.difference(self.available_goals)
            raise Exception(
                f"Unknown goal(s) {difference}, "
                f"following goals are avaialable {self.available_goals}"
            )

    def _check_if_build_steps_to_rebuild_are_valid_goals(self):
        build_steps_to_rebuild = self._get_build_steps_to_rebuild()
        if not build_steps_to_rebuild.issubset(self.available_goals):
            difference = build_steps_to_rebuild.difference(self.available_goals)
            raise Exception(
                f"Unknown build stages {difference} forced to rebuild, "
                f"following stages are avaialable {self.available_goals}"
            )

    def create_build_tasks(
        self, shortcut_build: bool = True
    ) -> dict[str, DockerCreateImageTask]:
        image_infos = {
            goal: analyze_task_future.get_output()
            for goal, analyze_task_future in self.analyze_tasks_futures.items()
        }
        tasks = self._create_build_tasks_for_image_infos(image_infos, shortcut_build)
        return tasks

    def _create_build_tasks_for_image_infos(
        self, image_infos: dict[str, ImageInfo], shortcut_build: bool
    ):
        result = {
            goal: self._create_build_task_for_image_info(image_info, shortcut_build)
            for goal, image_info in image_infos.items()
        }
        return result

    def _create_build_task_for_image_info(
        self, image_info: ImageInfo, shortcut_build: bool = True
    ) -> DockerCreateImageTask:
        if self._build_with_depenencies_is_requested(image_info, shortcut_build):
            task_for_image_info = self._create_build_task_with_dependencies(
                image_info, shortcut_build
            )
            return task_for_image_info
        else:
            image_name = f"{image_info.target_repository_name}:{image_info.target_tag}"
            task_for_image_info = self.create_child_task(
                DockerCreateImageTask, image_name=image_name, image_info=image_info
            )
            return task_for_image_info

    def _build_with_depenencies_is_requested(
        self, image_info: ImageInfo, shortcut_build: bool
    ):
        needs_to_be_build = image_info.image_state == ImageState.NEEDS_TO_BE_BUILD.name
        result = (
            (not shortcut_build or needs_to_be_build)
            and image_info.depends_on_images
            and len(image_info.depends_on_images) > 0
        )
        return result

    def _create_build_task_with_dependencies(
        self, image_info: ImageInfo, shortcut_build: bool = True
    ) -> DockerCreateImageTask:
        assert image_info.depends_on_images
        required_tasks = self._create_build_tasks_for_image_infos(
            image_info.depends_on_images, shortcut_build
        )
        required_task_infos = self._create_required_task_infos(required_tasks)
        image_info_copy = copy.copy(image_info)  # TODO looks not nice
        image_info_copy.depends_on_images = {}
        image_name = f"{image_info.target_repository_name}:{image_info.target_tag}"
        task_for_image_info = self.create_child_task(
            DockerCreateImageTaskWithDeps,
            image_name=image_name,
            image_info=image_info_copy,
            required_task_infos=required_task_infos,
        )
        return task_for_image_info

    def _create_required_task_infos(
        self, required_tasks: dict[str, DockerCreateImageTask]
    ) -> RequiredTaskInfoDict:
        result = {
            key: self._create_required_task_info(required_task)
            for key, required_task in required_tasks.items()
        }
        return RequiredTaskInfoDict(result)

    def _create_required_task_info(
        self, required_task: DockerCreateImageTask
    ) -> RequiredTaskInfo:
        required_task_info = RequiredTaskInfo(
            module_name=required_task.__module__,
            class_name=required_task.__class__.__name__,
            params=required_task.param_kwargs,
        )
        return required_task_info



import functools
import inspect
import os
import shlex
import shutil
import subprocess
import tempfile
from pathlib import Path
from sys import stderr
from typing import (
    Optional,
)

from exasol_integration_test_docker_environment.lib.docker import ContextDockerClient
from exasol_integration_test_docker_environment.lib.docker.container.utils import (
    remove_docker_container,
)
from exasol_integration_test_docker_environment.lib.docker.volumes.utils import (
    remove_docker_volumes,
)
from exasol_integration_test_docker_environment.lib.models.data.environment_info import (
    EnvironmentInfo,
)
from exasol_integration_test_docker_environment.lib.test_environment.ports import Ports
from exasol_integration_test_docker_environment.testing.docker_registry import (
    default_docker_repository_name,
)
from exasol_integration_test_docker_environment.testing.exaslct_docker_test_environment import (
    ExaslctDockerTestEnvironment,
)
from exasol_integration_test_docker_environment.testing.spawned_test_environments import (
    SpawnedTestEnvironments,
)
from exasol_integration_test_docker_environment.testing.utils import (
    check_db_version_from_env,
)


def _cleanup(env_name: str):
    remove_docker_container([f"test_container_{env_name}", f"db_container_{env_name}"])
    remove_docker_volumes([f"db_container_{env_name}_volume"])


def get_class(test_object):
    if test_object is None:
        return None
    # if test_object is a class then simply return it.
    # Otherwise test_object is an instance of a class, then
    # return its class.
    return test_object if inspect.isclass(test_object) else test_object.__class__


def get_test_flavor(test_class):
    if test_class is None:
        return None
    source_file_of_test_object = inspect.getsourcefile(test_class)
    return Path(os.path.realpath(source_file_of_test_object)).parent.joinpath(
        "resources/test-flavor"
    )


class ExaslctTestEnvironment:

    def __init__(
        self,
        test_object: object,
        executable: str = "./exaslct",
        clean_images_at_close: bool = True,
        name: Optional[str] = None,
    ) -> None:
        self.clean_images_at_close = clean_images_at_close
        self.executable = executable
        self.test_object = test_object
        self.test_class = get_class(test_object)
        self.flavor_path = get_test_flavor(self.test_class)
        self.name = name if name else self.test_class.__name__
        self._docker_repository_name = default_docker_repository_name(self.name)
        if "RUN_SLC_TESTS_WITHIN_CONTAINER" in os.environ:
            # We need to put the output directories into the workdir,
            # because only this is shared between the current container and
            # host. Only paths within this shared directory can be mounted
            # to docker container started by exaslct
            temp_dir_prefix_path = Path("./temp_outputs")
            temp_dir_prefix_path.mkdir(exist_ok=True)
            self.temp_dir = tempfile.mkdtemp(dir=temp_dir_prefix_path)
        else:
            self.temp_dir = tempfile.mkdtemp()
        self._update_attributes()

    @property
    def repository_name(self):
        return self._docker_repository_name

    @repository_name.setter
    def repository_name(self, value):
        self._docker_repository_name = value
        self._update_attributes()

    def _update_attributes(self):
        self.flavor_path_argument = f"--flavor-path {self.flavor_path}"
        repository_name = self.repository_name
        self.docker_repository_arguments = (
            f"--source-docker-repository-name {repository_name} "
            f"--target-docker-repository-name {repository_name}"
        )
        self.clean_docker_repository_arguments = (
            f"--docker-repository-name {repository_name}"
        )
        self.output_directory_arguments = f"--output-directory {self.temp_dir}"
        self.task_dependencies_argument = " ".join(
            [
                f"--task-dependencies-dot-file {self.name}.dot",
            ]
        )

    def clean_images(self):
        self.run_command(f"{self.executable} clean-flavor-images", clean=True)

    def run_command(
        self,
        command: str,
        use_output_directory: bool = True,
        use_flavor_path: bool = True,
        use_docker_repository: bool = True,
        track_task_dependencies: bool = False,
        clean: bool = False,
        capture_output: bool = False,
    ):
        if use_output_directory:
            command = f"{command} {self.output_directory_arguments}"
        if track_task_dependencies:
            command = f"{command} {self.task_dependencies_argument}"
        if use_flavor_path:
            command = f"{command} {self.flavor_path_argument}"
        if use_docker_repository and not clean:
            command = f"{command} {self.docker_repository_arguments}"
        if use_docker_repository and clean:
            command = f"{command} {self.clean_docker_repository_arguments}"
        print(file=stderr)
        print(f"command: {command}", file=stderr)
        if capture_output:
            completed_process = subprocess.run(
                shlex.split(command), stdout=subprocess.PIPE, stderr=subprocess.STDOUT
            )
        else:
            completed_process = subprocess.run(shlex.split(command))
        try:
            completed_process.check_returncode()
        except subprocess.CalledProcessError as e:
            if capture_output:
                print(e.stdout.decode("UTF-8"), file=stderr)
            raise e
        return completed_process

    def close(self):
        try:
            if self.clean_images_at_close:
                self.clean_images()
        except Exception as e:
            print(e, file=stderr)
        try:
            shutil.rmtree(self.temp_dir)
        except Exception as e:
            print(e, file=stderr)

    def spawn_docker_test_environments(
        self, name: str, additional_parameter: Optional[list[str]] = None
    ) -> SpawnedTestEnvironments:
        ports = Ports.random_free()
        on_host_parameter = ExaslctDockerTestEnvironment(
            name=self.name + "_" + name,
            database_host="localhost",
            db_username="sys",
            db_password="exasol",
            bucketfs_username="w",
            bucketfs_password="write",
            ports=ports,
        )

        arguments: list[str] = [
            f"--environment-name {on_host_parameter.name}",
            f"--database-port-forward {on_host_parameter.ports.database}",
            f"--bucketfs-port-forward {on_host_parameter.ports.bucketfs_http}",
            f"--bucketfs-https-port-forward {on_host_parameter.ports.bucketfs_https}",
            f"--ssh-port-forward {on_host_parameter.ports.ssh}",
        ]
        db_version = check_db_version_from_env()
        if db_version:
            arguments.append(f'--docker-db-image-version "{db_version}"')
        if additional_parameter:
            arguments += additional_parameter
        arguments_str = " ".join(arguments)

        command = f"{self.executable} spawn-test-environment {arguments_str}"
        completed_process = self.run_command(
            command,
            use_flavor_path=False,
            use_docker_repository=False,
            capture_output=True,
        )
        on_host_parameter.completed_process = completed_process
        environment_info_json_path = Path(
            self.temp_dir,
            f"cache/environments/{on_host_parameter.name}/environment_info.json",
        )
        if environment_info_json_path.exists():
            with environment_info_json_path.open() as f:
                environment_info = EnvironmentInfo.from_json(f.read())
                on_host_parameter.environment_info = environment_info
        on_host_parameter.clean_up = functools.partial(_cleanup, on_host_parameter.name)  # type: ignore
        if "RUN_SLC_TESTS_WITHIN_CONTAINER" in os.environ:
            slc_test_run_parameter = ExaslctDockerTestEnvironment(
                name=on_host_parameter.name,
                database_host="localhost",
                db_username=on_host_parameter.db_username,
                db_password=on_host_parameter.db_password,
                bucketfs_username=on_host_parameter.bucketfs_username,
                bucketfs_password=on_host_parameter.bucketfs_password,
                ports=Ports.default_ports,
                environment_info=on_host_parameter.environment_info,
                completed_process=on_host_parameter.completed_process,
            )

            with ContextDockerClient() as docker_client:
                db_container = docker_client.containers.get(
                    f"db_container_{slc_test_run_parameter.name}"
                )
                cloudbuild_network = docker_client.networks.get("cloudbuild")
                cloudbuild_network.connect(db_container)
                db_container.reload()
                slc_test_run_parameter.database_host = db_container.attrs[
                    "NetworkSettings"
                ]["Networks"][cloudbuild_network.name]["IPAddress"]
                return SpawnedTestEnvironments(
                    on_host_parameter, slc_test_run_parameter
                )
        else:
            return SpawnedTestEnvironments(on_host_parameter, None)



import luigi

from exasol_integration_test_docker_environment.lib.models.data.environment_type import (
    EnvironmentType,
)
from exasol_integration_test_docker_environment.lib.test_environment.parameter.docker_db_test_environment_parameter import (
    DockerDBTestEnvironmentParameter,
)
from exasol_integration_test_docker_environment.lib.test_environment.parameter.external_test_environment_parameter import (
    ExternalDatabaseCredentialsParameter,
)
from exasol_integration_test_docker_environment.lib.test_environment.parameter.general_spawn_test_environment_parameter import (
    GeneralSpawnTestEnvironmentParameter,
)


class SpawnTestEnvironmentParameter(
    GeneralSpawnTestEnvironmentParameter,
    ExternalDatabaseCredentialsParameter,
    DockerDBTestEnvironmentParameter,
):
    environment_type = luigi.EnumParameter(enum=EnvironmentType)



from typing import Optional

from exasol_integration_test_docker_environment.lib.base.info import Info
from exasol_integration_test_docker_environment.lib.models.data.container_info import (
    ContainerInfo,
)
from exasol_integration_test_docker_environment.lib.models.data.ssh_info import SshInfo
from exasol_integration_test_docker_environment.lib.test_environment.ports import Ports


class DatabaseInfo(Info):
    def __init__(
        self,
        host: str,
        ports: Ports,
        reused: bool,
        container_info: Optional[ContainerInfo] = None,
        ssh_info: Optional[SshInfo] = None,
        forwarded_ports: Optional[Ports] = None,
    ) -> None:
        self.container_info = container_info
        self.ports = ports
        self.host = host
        self.reused = reused
        self.ssh_info = ssh_info
        self.forwarded_ports = forwarded_ports



import pathlib
from datetime import (
    datetime,
    timezone,
)

from exasol_integration_test_docker_environment.lib.logging.abstract_log_handler import (
    AbstractLogHandler,
)
from exasol_integration_test_docker_environment.lib.models.config.log_config import (
    WriteLogFilesToConsole,
)


class CommandLogHandler(AbstractLogHandler):

    def __init__(self, log_file_path: pathlib.Path, logger, description: str):
        super().__init__(log_file_path, logger)
        self._description = description

    def handle_log_line(self, log_line, error: bool = False):
        cur_time = datetime.now(timezone.utc)
        log_time = cur_time.strftime("%H.%M.%S.%f")
        self._log_file.write(f"{log_time}: {log_line}\n")
        self._log_file.flush()
        self._complete_log.append(f"{log_time}: {log_line}\n")
        if error:
            self._error_message = log_line

    def finish(self):
        if self._log_config.write_log_files_to_console == WriteLogFilesToConsole.all:
            self._logger.info(
                "Command log for %s \n%s",
                self._description,
                "".join(self._complete_log),
            )
        if self._error_message is not None:
            if (
                self._log_config.write_log_files_to_console
                == WriteLogFilesToConsole.only_error
            ):
                self._logger.error(
                    "Command failed %s failed\nCommand Log:\n%s",
                    self._description,
                    "\n".join(self._complete_log),
                )
            raise Exception(
                'Error occurred during %s. Received error "%s" .'
                "The whole log can be found in %s"
                % (
                    self._description,
                    self._error_message,
                    self._log_file_path.absolute(),
                ),
                self._log_file_path.absolute(),
            )



from typing import Optional

from exasol_integration_test_docker_environment.testing.exaslct_docker_test_environment import (
    ExaslctDockerTestEnvironment,
)


class SpawnedTestEnvironments:
    def __init__(
        self,
        on_host_environment: ExaslctDockerTestEnvironment,
        slc_test_run_environment: Optional[ExaslctDockerTestEnvironment],
    ) -> None:
        self.on_host_docker_environment = on_host_environment
        self.slc_test_run_environment = slc_test_run_environment

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        if self.on_host_docker_environment is not None:
            self.on_host_docker_environment.close()

        if self.slc_test_run_environment is not None:
            self.slc_test_run_environment.close()



import abc
from pathlib import Path

import docker
import git

from exasol_integration_test_docker_environment.abstract_method_exception import (
    AbstractMethodException,
)
from exasol_integration_test_docker_environment.lib.base.docker_base_task import (
    DockerBaseTask,
)
from exasol_integration_test_docker_environment.lib.docker.images.create.utils.build_context_hasher import (
    BuildContextHasher,
)
from exasol_integration_test_docker_environment.lib.docker.images.create.utils.docker_image_target import (
    DockerImageTarget,
)
from exasol_integration_test_docker_environment.lib.docker.images.create.utils.docker_registry_image_checker import (
    DockerRegistryImageChecker,
)
from exasol_integration_test_docker_environment.lib.docker.images.image_info import (
    ImageDescription,
    ImageInfo,
    ImageState,
)
from exasol_integration_test_docker_environment.lib.models.config.build_config import (
    build_config,
)
from exasol_integration_test_docker_environment.lib.models.config.docker_config import (
    docker_build_arguments,
)


class DockerAnalyzeImageTask(DockerBaseTask):

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._source_repository_name = self.get_source_repository_name()
        self._target_repository_name = self.get_target_repository_name()
        self._source_image_tag = self.get_source_image_tag()
        self._target_image_tag = self.get_target_image_tag()
        merged_transparent_build_arguments = {
            **self.get_transparent_build_arguments(),
            **docker_build_arguments().transparent,
        }
        merged_image_changing_build_arguments = {
            **self.get_image_changing_build_arguments(),
            **docker_build_arguments().image_changing,
        }
        self.image_description = ImageDescription(
            dockerfile=self.get_dockerfile(),
            mapping_of_build_files_and_directories=self.get_mapping_of_build_files_and_directories(),
            image_changing_build_arguments=merged_image_changing_build_arguments,
            transparent_build_arguments=merged_transparent_build_arguments,
        )
        self._dockerfile = self.get_dockerfile()

    @abc.abstractmethod
    def get_source_repository_name(self) -> str:
        """
        Called by the constructor to get the image name for pulls. Sub classes need to implement this method.
        :return: image name
        """
        raise AbstractMethodException()

    @abc.abstractmethod
    def get_target_repository_name(self) -> str:
        """
        Called by the constructor to get the image name for pushs. Sub classes need to implement this method.
        :return: image name
        """
        raise AbstractMethodException()

    @abc.abstractmethod
    def get_source_image_tag(self) -> str:
        """
        Called by the constructor to get the image tag for pulls. Sub classes need to implement this method.
        :return: image tag
        """
        raise AbstractMethodException()

    @abc.abstractmethod
    def get_target_image_tag(self) -> str:
        """
        Called by the constructor to get the image tag for pushs. Sub classes need to implement this method.
        :return: image tag
        """
        raise AbstractMethodException()

    def get_mapping_of_build_files_and_directories(self) -> dict[str, str]:
        """
        Called by the constructor to get the build files and directories mapping.
        The keys are the relative paths to the destination in build context and
        the values are the paths to the source directories or files.
        Sub classes need to implement this method.
        :return: dictionaries with destination path as keys and source paths in values
        """
        raise AbstractMethodException()

    @abc.abstractmethod
    def get_dockerfile(self) -> str:
        """
        Called by the constructor to get the path to the dockerfile.
        Sub classes need to implement this method.
        :return: path to the dockerfile
        """
        raise AbstractMethodException()

    def get_image_changing_build_arguments(self) -> dict[str, str]:
        """
        Called by the constructor to get the path image changing docker build arguments.
        Different values for these arguments might change the image, such that they
        will be part of the image hash and can cause a build.
        A common use case is define a mirror from which packages should be installed.
        Sub classes need to implement this method.
        :return: Dictionary of build arguments, where the keys are the argument name
        """
        return {}

    def get_transparent_build_arguments(self) -> dict[str, str]:
        """
        Called by the constructor to get the path transparent docker build arguments.
        Transparent arguments do not change the contain of the images.
        They are not part of the image hash. A common use case is define a mirror
        from which packages should be installed.
        Sub classes need to implement this method.
        :return: Dictionary of build arguments, where the keys are the argument name
        """
        return {}

    @abc.abstractmethod
    def is_rebuild_requested(self) -> bool:
        raise AbstractMethodException()

    def register_required(self):
        task_classes = self.requires_tasks()
        if task_classes is not None:
            if (
                not isinstance(task_classes, dict)
                or not self.keys_are_string(task_classes)
                or not self.values_are_subclass_of_baseclass(task_classes)
            ):
                raise TypeError(
                    f"Expected Dict[str,DockerAnalyzeImageTask] got {task_classes}"
                )
            tasks = {
                key: self.create_child_task_with_common_params(value)
                for key, value in task_classes.items()
            }
        else:
            tasks = None
        self.dependencies_futures = self.register_dependencies(tasks)

    def keys_are_string(self, task_classes) -> bool:
        return all(isinstance(key, str) for key in task_classes.keys())

    def values_are_subclass_of_baseclass(self, task_classes) -> bool:
        return all(
            issubclass(value, DockerAnalyzeImageTask) for value in task_classes.values()
        )

    def requires_tasks(self) -> dict[str, type["DockerAnalyzeImageTask"]]:
        return {}

    def run_task(self) -> None:
        image_info_of_dependencies = self.get_values_from_futures(
            self.dependencies_futures
        )
        if image_info_of_dependencies is None:
            image_info_of_dependencies = {}
        _build_context_hasher = BuildContextHasher(self.logger, self.image_description)
        image_hash = _build_context_hasher.generate_image_hash(
            image_info_of_dependencies
        )
        build_name = build_config().build_name
        image_info = ImageInfo(
            source_repository_name=self._source_repository_name,
            target_repository_name=self._target_repository_name,
            source_tag=self._source_image_tag,
            target_tag=self._target_image_tag,
            hash_value=image_hash,
            commit=self.get_commit_id(),
            build_name=build_name if build_name else "",
            depends_on_images=image_info_of_dependencies,
            image_state=None,
            image_description=self.image_description,
        )
        target_image_target = DockerImageTarget(
            self._target_repository_name, image_info.get_target_complete_tag()
        )
        source_image_target = DockerImageTarget(
            self._source_repository_name, image_info.get_source_complete_tag()
        )
        image_state = self.get_image_state(
            source_image_target, target_image_target, image_info_of_dependencies
        )
        image_info.image_state = image_state.name  # TODO setter for image_state
        self.return_object(image_info)

    def get_commit_id(self):
        try:
            repo = git.Repo(search_parent_directories=True)
            sha = repo.head.object.hexsha
            return sha
        except Exception as e:
            self.logger.info(
                "Not a Git Repository, can't determine the commit id for the image_info"
            )
            return ""

    def get_image_state(
        self,
        source_image_target: DockerImageTarget,
        target_image_target: DockerImageTarget,
        image_info_of_dependencies: dict[str, ImageInfo],
    ) -> ImageState:

        if self.is_rebuild_necessary(image_info_of_dependencies):
            return ImageState.NEEDS_TO_BE_BUILD
        else:
            if (
                self.is_image_locally_available(target_image_target)
                and not build_config().force_pull
                and not build_config().force_load
            ):
                return ImageState.TARGET_LOCALLY_AVAILABLE
            if (
                self.is_image_locally_available(source_image_target)
                and not build_config().force_pull
                and not build_config().force_load
            ):
                return ImageState.SOURCE_LOCALLY_AVAILABLE
            elif self.can_image_be_loaded(source_image_target):
                return ImageState.CAN_BE_LOADED
            elif self.is_image_in_registry(source_image_target):
                return ImageState.REMOTE_AVAILABLE
            else:
                return ImageState.NEEDS_TO_BE_BUILD

    def needs_any_dependency_to_be_build(
        self, image_info_of_dependencies: dict[str, ImageInfo]
    ) -> bool:
        return any(
            image_info.image_state == ImageState.NEEDS_TO_BE_BUILD
            for image_info in image_info_of_dependencies.values()
        )

    def is_rebuild_necessary(self, image_info_of_dependencies: dict[str, ImageInfo]):
        return (
            self.needs_any_dependency_to_be_build(image_info_of_dependencies)
            or self.is_rebuild_requested()
        )

    def is_image_locally_available(self, image_target: DockerImageTarget):
        exists = image_target.exists()
        self.logger.info(
            f"Checking if image {image_target.get_complete_name()} "
            f"is locally available, result {exists}"
        )
        return exists

    def can_image_be_loaded(self, image_target: DockerImageTarget):
        if build_config().cache_directory is not None:
            image_path = self.get_path_to_cached_image(image_target)
            self.logger.info(
                f"Checking if image archive {image_path} "
                f"is available in cache directory, "
                f"result {image_path.exists()}"
            )
            return image_path.exists()
        else:
            return False

    def get_path_to_cached_image(self, image_target):
        image_path = Path(build_config().cache_directory).joinpath(
            Path(image_target.get_complete_name() + ".tar")
        )
        return image_path

    def is_image_in_registry(self, image_target: DockerImageTarget):
        try:
            self.logger.info(
                "Try to find image %s in registry", image_target.get_complete_name()
            )
            exists = DockerRegistryImageChecker().check(
                image_target.get_complete_name()
            )
            if exists:
                self.logger.info(
                    "Found image %s in registry", image_target.get_complete_name()
                )
            return exists
        except docker.errors.DockerException as e:
            self.logger.warning(
                "Image %s not in registry, got exception %s",
                image_target.get_complete_name(),
                e,
            )
            return False



from pathlib import Path

from exasol_integration_test_docker_environment.lib.docker.images.create.docker_image_creator_base_task import (
    DockerImageCreatorBaseTask,
)
from exasol_integration_test_docker_environment.lib.models.config.build_config import (
    build_config,
)


class DockerLoadImageTask(DockerImageCreatorBaseTask):

    def run_task(self) -> None:
        cache_directory = build_config().cache_directory or ""
        image_archive_path = Path(cache_directory).joinpath(
            self.image_info.get_source_complete_name() + ".tar"
        )
        self.logger.info(
            "Try to load docker image %s from %s",
            self.image_info.get_source_complete_name(),
            image_archive_path,
        )
        with self._get_docker_client() as docker_client:
            with image_archive_path.open("rb") as f:
                docker_client.images.load(f)
            docker_client.images.get(self.image_info.get_source_complete_name()).tag(
                repository=self.image_info.target_repository_name,
                tag=self.image_info.get_target_complete_tag(),
            )



import sys

from exasol_integration_test_docker_environment.cli.cli import cli
from exasol_integration_test_docker_environment.lib import api
from exasol_integration_test_docker_environment.lib.models.api_errors import (
    HealthProblem,
)


@cli.command()
def health():
    """
    Check the health of the execution environment.

    If no issues have been found, using the library or executing the test should work just fine.
    For all found issues there will be a proposed fix/solution.

    If the environment was found to be healthy the exit code will be 0, otherwise -1.
    """
    success, failure = 0, -1
    try:
        api.health()
        sys.exit(success)
    except HealthProblem as e:
        print(e.args[0])
        sys.exit(failure)



from collections.abc import (
    Generator,
    Iterator,
)
from typing import (
    Optional,
)

import docker
import luigi

import exasol_integration_test_docker_environment.certificate_resources.container
from exasol_integration_test_docker_environment.lib.base.docker_base_task import (
    DockerBaseTask,
)
from exasol_integration_test_docker_environment.lib.docker.images.image_info import (
    ImageInfo,
)
from exasol_integration_test_docker_environment.lib.models.data.docker_volume_info import (
    DockerVolumeInfo,
)
from exasol_integration_test_docker_environment.lib.test_environment.create_certificates.analyze_certificate_container import (
    DockerCertificateBuildBase,
    DockerCertificateContainerBuild,
)
from exasol_integration_test_docker_environment.lib.utils.resource_directory import (
    ResourceDirectory,
)

CERTIFICATES_MOUNT_PATH = "/certificates"


class CreateSSLCertificatesTask(DockerBaseTask):
    environment_name: str = luigi.Parameter()
    docker_runtime: Optional[str] = luigi.OptionalParameter(
        default=None, significant=False
    )
    db_container_name: str = luigi.Parameter(significant=False)
    network_name: str = luigi.Parameter()
    reuse: bool = luigi.BoolParameter(default=False, significant=False)
    no_cleanup_after_success: bool = luigi.BoolParameter(
        default=False, significant=False
    )
    no_cleanup_after_failure: bool = luigi.BoolParameter(
        default=False, significant=False
    )
    volume_name: str = luigi.Parameter()

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._temp_resource_directory = ResourceDirectory(
            exasol_integration_test_docker_environment.certificate_resources.container
        )
        self.volume_info: Optional[DockerVolumeInfo] = None
        self._temp_resource_directory.create()

    def on_failure(self, exception) -> None:
        super().on_failure(exception)
        self._temp_resource_directory.cleanup()

    def on_success(self) -> None:
        super().on_success()
        self._temp_resource_directory.cleanup()

    def run_task(self) -> Iterator[DockerCertificateContainerBuild]:
        self.volume_info = None
        image_infos = yield from self.build_image()
        if self.reuse:
            self.logger.info("Try to reuse volume %s", self.volume_name)
            try:
                self.volume_info = self.get_volume_info(reused=True)
            except Exception as e:
                self.logger.warning(
                    f"Tried to reuse volume {self.volume_name}, but got Exeception {e}. "
                    "Fallback to create new volume."
                )
        if self.volume_info is None:
            self.volume_info = self.create_docker_volume()
            self.create_certificate(image_infos, self.volume_info)

        self.return_object(self.volume_info)

    def build_image(
        self,
    ) -> Generator[DockerCertificateContainerBuild, None, dict[str, ImageInfo]]:
        task = self.create_child_task(
            task_class=DockerCertificateContainerBuild,
            certificate_container_root_directory=self._temp_resource_directory.tmp_directory,
        )
        image_infos_future = yield from self.run_dependencies(task)
        image_infos: dict[str, ImageInfo] = self.get_values_from_future(image_infos_future)  # type: ignore
        return image_infos

    def get_volume_info(self, reused: bool) -> DockerVolumeInfo:
        with self._get_docker_client() as docker_client:
            volume_properties = docker_client.api.inspect_volume(self.volume_name)
            if volume_properties["Name"] == self.volume_name:
                return DockerVolumeInfo(
                    volume_name=str(self.volume_name),
                    mount_point=volume_properties["Mountpoint"],
                    reused=reused,
                )
        raise RuntimeError(
            f"Volume Info not found for created volume {self.volume_name}"
        )

    def create_docker_volume(self) -> DockerVolumeInfo:
        self.remove_volume(self.volume_name)
        with self._get_docker_client() as docker_client:
            volume = docker_client.volumes.create(
                name=self.volume_name,
            )
            volume_info = self.get_volume_info(reused=False)
        return volume_info

    def remove_volume(self, volume_name):
        try:
            with self._get_docker_client() as docker_client:
                docker_client.volumes.get(volume_name).remove()
                self.logger.info("Removed volume %s", volume_name)
        except docker.errors.NotFound:
            pass

    def cleanup_task(self, success):
        if (success and not self.no_cleanup_after_success) or (
            not success and not self.no_cleanup_after_failure
        ):
            try:
                self.logger.info(f"Cleaning up volume %s:", self.volume_name)
                self.remove_volume(self.volume_name)
            except Exception as e:
                self.logger.error(
                    f"Error during removing volume %s: %s:", self.volume_name, e
                )

    @property
    def _construct_complete_host_name(self):
        """
        Example: 'db_container_exasol_test.db_network_exasol_test'
        """
        return f"{self.db_container_name}.{self.network_name}"

    def create_certificate(
        self, image_infos: dict[str, ImageInfo], volume_info: DockerVolumeInfo
    ) -> None:
        certificate_container_image_info = image_infos[DockerCertificateBuildBase.GOAL]
        volumes = {
            volume_info.volume_name: {
                "bind": CERTIFICATES_MOUNT_PATH,
                "mode": "rw",
            }
        }

        with self._get_docker_client() as docker_client:
            try:
                container = docker_client.containers.create(
                    image=certificate_container_image_info.get_target_complete_name(),
                    name="certificate_resources",
                    network_mode=None,
                    command="sleep infinity",
                    detach=True,
                    volumes=volumes,
                    labels={
                        "test_environment_name": self.environment_name,
                        "container_type": "certificate_resources",
                    },
                    runtime=self.docker_runtime,
                )
                container.start()
                self.logger.info("Creating certificates...")
                cmd = (
                    f"bash /scripts/create_certificates.sh "
                    f"{self._construct_complete_host_name} {CERTIFICATES_MOUNT_PATH}"
                )
                exit_code, output = container.exec_run(cmd)
                self.logger.info(output.decode("utf-8"))
                if exit_code != 0:
                    raise RuntimeError(
                        f"Error creating certificates:'{output.decode('utf-8')}'"
                    )
            finally:
                container.stop()
                container.remove()



import time
from abc import abstractmethod
from typing import (
    Optional,
    Protocol,
    runtime_checkable,
)

import fabric
from docker import DockerClient
from docker.models.containers import (
    Container,
    ExecResult,
)
from paramiko.ssh_exception import NoValidConnectionsError

from exasol_integration_test_docker_environment.lib.base.ssh_access import SshKey
from exasol_integration_test_docker_environment.lib.docker import ContextDockerClient
from exasol_integration_test_docker_environment.lib.models.data.database_info import (
    DatabaseInfo,
)


class DockerClientFactory:
    """
    Create a Docker client.
    """

    def __init__(self, timeout: int = 100000) -> None:
        self._timeout = timeout

    def client(self) -> DockerClient:
        with ContextDockerClient(timeout=self._timeout) as client:
            return client


# Avoid TypeError: Instance and class checks can only be
# used with @runtime_checkable protocols
# raised by unit tests
@runtime_checkable
class DbOsExecutor(Protocol):
    """
    This class provides an abstraction to execute operating system
    commands on the database host, e.g. inside a Docker Container.  See
    concrete implementations in sub-classes ``DockerExecutor`` and
    ``SshExecutor``.
    """

    def exec(self, cmd: str) -> ExecResult: ...

    def prepare(self): ...


class DockerExecutor(DbOsExecutor):
    def __init__(self, docker_client: DockerClient, container_name: str) -> None:
        self._client = docker_client
        self._container_name = container_name
        self._container: Optional[Container] = None

    def __enter__(self):
        self._container = self._client.containers.get(self._container_name)
        return self

    def __exit__(self, type_, value, traceback):
        self.close()

    def __del__(self):
        self.close()

    def exec(self, cmd: str) -> ExecResult:
        assert self._container
        return self._container.exec_run(cmd)

    def prepare(self):
        pass

    def close(self):
        self._container = None
        if self._client is not None:
            self._client.close()
            self._client = None


class SshExecutor(DbOsExecutor):
    def __init__(self, connect_string: str, key_file: str) -> None:
        self._connect_string = connect_string
        self._key_file = key_file
        self._connection: Optional[fabric.Connection] = None

    def __enter__(self):
        key = SshKey.read_from(self._key_file)
        self._connection = fabric.Connection(
            self._connect_string,
            connect_kwargs={"pkey": key.private},
        )
        return self

    def __exit__(self, type_, value, traceback):
        self.close()

    def __del__(self):
        self.close()

    def exec(self, cmd: str) -> ExecResult:
        assert self._connection
        result = self._connection.run(cmd, warn=True, hide=True)
        output = result.stdout.encode("utf-8")
        return ExecResult(result.exited, output)

    def prepare(self):
        retry = 0
        while retry < 20:
            try:
                retry += 1
                self._connection.run("true", warn=True, hide=True)
                break
            except NoValidConnectionsError as ex:
                if retry > 20:
                    raise ex
                time.sleep(1)

    def close(self):
        if self._connection is not None:
            self._connection.close()
            self._connection = None


# Avoid TypeError: Instance and class checks can only be
# used with @runtime_checkable protocols
# raised by integration tests
@runtime_checkable
class DbOsExecFactory(Protocol):
    """
    This class defines an abstract method ``executor()`` to be implemented by
    inheriting factories.
    """

    @abstractmethod
    def executor(self) -> DbOsExecutor:
        """
        Create an executor for executing commands inside of the operating
        system of the database.
        """
        ...


class DockerExecFactory(DbOsExecFactory):
    def __init__(
        self, container_name: str, client_factory: DockerClientFactory
    ) -> None:
        self._container_name = container_name
        self._client_factory = client_factory

    def executor(self) -> DbOsExecutor:
        client = self._client_factory.client()
        return DockerExecutor(client, self._container_name)


class SshExecFactory(DbOsExecFactory):
    @classmethod
    def from_database_info(cls, info: DatabaseInfo):
        assert info.ssh_info
        return SshExecFactory(
            f"{info.ssh_info.user}@{info.host}:{info.ports.ssh}",
            info.ssh_info.key_file,
        )

    def __init__(self, connect_string: str, ssh_key_file: str) -> None:
        self._connect_string = connect_string
        self._key_file = ssh_key_file

    def executor(self) -> DbOsExecutor:
        return SshExecutor(self._connect_string, self._key_file)



import contextlib
from collections.abc import Iterator
from typing import (
    Callable,
    ContextManager,
    Optional,
)

from typing_extensions import TypeAlias  # Needed for Python3.9

from exasol_integration_test_docker_environment.testing import utils
from exasol_integration_test_docker_environment.testing.exaslct_test_environment import (
    ExaslctTestEnvironment,
)
from exasol_integration_test_docker_environment.testing.spawned_test_environments import (
    SpawnedTestEnvironments,
)

CliContextProvider: TypeAlias = Callable[
    [Optional[str], Optional[list[str]]], ContextManager[SpawnedTestEnvironments]
]


@contextlib.contextmanager
def build_cli_isolation(request) -> Iterator[ExaslctTestEnvironment]:
    """
    Builds an ExaslctTestEnvironment instance with a proper name based on the pytest request fixture.
    Cleans up the environment automatically on shutdown.
    """
    testname = utils.normalize_request_name(request.node.name)
    environment = ExaslctTestEnvironment(
        test_object=None,
        executable="itde",
        clean_images_at_close=True,
        name=testname,
    )
    yield environment
    utils.close_environments(environment)


def build_cli_context_provider(
    test_environment: ExaslctTestEnvironment,
) -> CliContextProvider:
    """
    Returns a context provider function which can be used to spawn a Docker DB with custom name
    and custom additional db parameters.
    Cleans up the docker db automatically on shutdown when context provider closes.
    """

    @contextlib.contextmanager
    def create_context(
        name: Optional[str] = None,
        additional_parameters: Optional[list[str]] = None,
    ) -> Iterator[SpawnedTestEnvironments]:
        name = name if name else test_environment.name
        spawned = test_environment.spawn_docker_test_environments(
            name=name,
            additional_parameter=additional_parameters,
        )
        yield spawned
        utils.close_environments(spawned)

    return create_context



from collections.abc import Iterator

import luigi

from exasol_integration_test_docker_environment.lib.base.base_task import BaseTaskType
from exasol_integration_test_docker_environment.lib.base.docker_base_task import (
    DockerBaseTask,
)
from exasol_integration_test_docker_environment.lib.docker.images.utils import (
    find_images_by_tag,
)


class CleanImageTask(DockerBaseTask):
    image_id: str = luigi.Parameter()

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def run_task(self) -> Iterator["CleanImageTask"]:
        self.logger.info("Try to remove dependent images of %s" % self.image_id)
        yield from self.run_dependencies(
            self.get_clean_image_tasks_for_dependent_images()
        )
        for i in range(3):
            try:
                self.logger.info("Try to remove image %s" % self.image_id)
                with self._get_docker_client() as docker_client:
                    docker_client.images.remove(image=self.image_id, force=True)
                self.logger.info("Removed image %s" % self.image_id)
                break
            except Exception as e:
                self.logger.info(
                    "Could not removed image {} got exception {}".format(
                        self.image_id, e
                    )
                )

    def get_clean_image_tasks_for_dependent_images(self) -> list["CleanImageTask"]:
        with self._get_docker_client() as docker_client:
            image_ids = [
                str(possible_child).replace("sha256:", "")
                for possible_child in docker_client.api.images(all=True, quiet=True)
                if self.is_child_image(possible_child)
            ]
            return [
                self.create_child_task(CleanImageTask, image_id=image_id)
                for image_id in image_ids
            ]

    def is_child_image(self, possible_child):
        try:
            with self._get_docker_client() as docker_client:
                inspect = docker_client.api.inspect_image(
                    image=str(possible_child).replace("sha256:", "")
                )
            return str(inspect["Parent"]).replace("sha256:", "") == self.image_id
        except:
            return False


class CleanImagesStartingWith(DockerBaseTask):

    starts_with_pattern: str = luigi.Parameter()

    def register_required(self):
        image_ids = [
            str(image.id).replace("sha256:", "")
            for image in self.find_images_to_clean()
        ]
        self.register_dependencies(
            [
                self.create_child_task(CleanImageTask, image_id=image_id)
                for image_id in image_ids
            ]
        )

    def find_images_to_clean(self):
        self.logger.info(
            "Going to remove all images starting with %s" % self.starts_with_pattern
        )
        with self._get_docker_client() as docker_client:
            filter_images = find_images_by_tag(
                docker_client, lambda tag: tag.startswith(self.starts_with_pattern)
            )
            for i in filter_images:
                self.logger.info("Going to remove following image: %s" % i.tags)
            return filter_images

    def run_task(self) -> None:
        pass



from typing import Optional

import luigi
from luigi import Config
from luigi.parameter import ParameterVisibility


class ExternalDatabaseXMLRPCParameter(Config):
    external_exasol_xmlrpc_host: Optional[str] = luigi.OptionalParameter()
    external_exasol_xmlrpc_port: int = luigi.IntParameter(default=443)
    external_exasol_xmlrpc_user: Optional[str] = luigi.OptionalParameter()
    external_exasol_xmlrpc_cluster_name: Optional[str] = luigi.OptionalParameter()
    external_exasol_xmlrpc_password: Optional[str] = luigi.OptionalParameter(
        significant=False,
        visibility=ParameterVisibility.HIDDEN,
    )


# See ticket https://github.com/exasol/integration-test-docker-environment/issues/341
class ExternalDatabaseHostParameter(Config):
    external_exasol_db_host: Optional[str] = luigi.OptionalParameter()
    external_exasol_db_port: int = luigi.IntParameter()
    external_exasol_bucketfs_http_port: int = luigi.IntParameter()
    external_exasol_bucketfs_https_port: int = luigi.IntParameter()
    external_exasol_ssh_port: int = luigi.IntParameter()


class ExternalDatabaseCredentialsParameter(
    ExternalDatabaseHostParameter,
    ExternalDatabaseXMLRPCParameter,
):
    external_exasol_db_user: Optional[str] = luigi.OptionalParameter()
    external_exasol_db_password: Optional[str] = luigi.OptionalParameter(
        significant=False,
        visibility=ParameterVisibility.HIDDEN,
    )
    external_exasol_bucketfs_write_password: Optional[str] = luigi.OptionalParameter(
        significant=False,
        visibility=ParameterVisibility.HIDDEN,
    )



import click


@click.group()
def cli():
    """
    ITDE - Integration Test Docker Environment

    Create and control a docker based exasol database test setup.

    Examples:

        Check the health of the execution environment:

            $ itde health

        Spawn an ITDE test environment:

            $ itde spawn-test-environment --environment-name test \\
            --database-port-forward 8563 --bucketfs-port-forward 2580 \\
            --docker-db-image-version 7.1.9 --db-mem-size 4GB

        Display the default configurations of the execution environment:

            $ itde environment --show-default-db-version
    """
    pass



from enum import (
    Enum,
    auto,
)
from typing import (
    Optional,
)

import luigi
from luigi import Config


class DbOsAccess(Enum):
    """
    This enum represents different methods to access the operating system
    of the Docker Container running the Exasol database.

    The Docker Container of older database versions allowed to use
    ``docker_exec`` as well, while newer versions require SSH.

    Currently the SSH access is not implemented fully but the current test
    case will verify that when using SSH access a file with the required
    private key is generated.
    """

    DOCKER_EXEC = auto()
    SSH = auto()


class DockerDBTestEnvironmentParameter(Config):
    docker_db_image_name: Optional[str] = luigi.OptionalParameter(default=None)
    docker_db_image_version: Optional[str] = luigi.OptionalParameter(default=None)
    reuse_database: bool = luigi.BoolParameter(default=False, significant=False)
    db_os_access = luigi.EnumParameter(
        default=DbOsAccess.DOCKER_EXEC, enum=DbOsAccess, significant=False
    )
    no_database_cleanup_after_success: bool = luigi.BoolParameter(
        default=False, significant=False
    )
    no_database_cleanup_after_failure: bool = luigi.BoolParameter(
        default=False, significant=False
    )
    database_port_forward: Optional[str] = luigi.OptionalParameter(
        default=None, significant=False
    )
    bucketfs_port_forward: Optional[str] = luigi.OptionalParameter(
        default=None, significant=False
    )
    bucketfs_http_port_forward: Optional[str] = luigi.OptionalParameter(
        default=None, significant=False
    )
    bucketfs_https_port_forward: Optional[str] = luigi.OptionalParameter(
        default=None, significant=False
    )
    ssh_port_forward: Optional[str] = luigi.OptionalParameter(
        default=None, significant=False
    )
    mem_size: Optional[str] = luigi.OptionalParameter(
        default="2 GiB", significant=False
    )
    disk_size: Optional[str] = luigi.OptionalParameter(
        default="2 GiB", significant=False
    )
    nameservers: list[str] = luigi.ListParameter(default=[], significant=False)



from pathlib import PurePath


class CharacterLengthChecker:
    """
    This helper class counts the number of character of directory- and file-names.
    If this number exceeds a given limit it raises an exception.
    The goal is to avoid high memory consumption for calculation of huge directories
    which contain many sub-directories/files (broad directory trees) and long names.
    """

    def __init__(
        self,
        root_directory: PurePath,
        max_characters_paths: int,
        count_directory_names: bool,
        count_file_names: bool,
    ) -> None:
        self._num_characters = 0
        self._max_characters_paths = max_characters_paths
        self._root_directory = root_directory
        self._count_directory_names = count_directory_names
        self._count_file_names = count_file_names

    def add_and_check(self, directories: list[str], files: list[str]) -> None:
        """
        Adds number of characters of the names of all directories and files to the internal counter
        and applies the check if the internal counter becomes greater than the given limit.
        :param directories Characters of all names of this list will be counted
        :param files Characters of all names of this list will be counted
        raises: OSError if the counted character is greater than the given limit
        """
        if self._count_directory_names:
            self._num_characters += sum([len(d) for d in directories])
        if self._count_file_names:
            self._num_characters += sum([len(f) for f in files])

        if self._num_characters > self._max_characters_paths:
            raise OSError(
                f"Walking through too many directories. Aborting. Please verify: {self._root_directory}"
            )



import contextlib
import ssl
import subprocess
from collections.abc import Iterator
from typing import Optional

import pyexasol
from pyexasol import ExaConnection

from exasol_integration_test_docker_environment.cli.options.test_environment_options import (
    LATEST_DB_VERSION,
)
from exasol_integration_test_docker_environment.lib.models.data.environment_info import (
    EnvironmentInfo,
)
from exasol_integration_test_docker_environment.lib.test_environment.ports import Ports
from exasol_integration_test_docker_environment.testing.utils import (
    check_db_version_from_env,
)


class ExaslctDockerTestEnvironment:
    def __init__(
        self,
        name: str,
        database_host: str,
        db_username: str,
        db_password: str,
        bucketfs_username: str,
        bucketfs_password: str,
        ports: Ports,
        environment_info: Optional[EnvironmentInfo] = None,
        completed_process: Optional[subprocess.CompletedProcess] = None,
    ) -> None:
        self.db_password = db_password
        self.db_username = db_username
        self.ports = ports
        self.bucketfs_username = bucketfs_username
        self.bucketfs_password = bucketfs_password
        self.database_host = database_host
        self.name = name
        self.environment_info = environment_info
        self.completed_process = completed_process
        self.clean_up = None
        self.docker_db_image_version = check_db_version_from_env() or LATEST_DB_VERSION

    def close(self):
        if self.clean_up is not None:
            self.clean_up()

    @contextlib.contextmanager
    def create_connection(self) -> Iterator[ExaConnection]:
        host_name = self.database_host
        port = self.ports.database
        dsn = f"{host_name}:{port}"
        connection = pyexasol.connect(
            dsn=dsn,
            user="sys",
            password="exasol",
            websocket_sslopt={"cert_reqs": ssl.CERT_NONE},
        )
        yield connection
        connection.close()



import os
import pathlib
import re
from collections.abc import Generator

import luigi

from exasol_integration_test_docker_environment.abstract_method_exception import (
    AbstractMethodException,
)
from exasol_integration_test_docker_environment.lib.base.docker_base_task import (
    DockerBaseTask,
)
from exasol_integration_test_docker_environment.lib.base.still_running_logger import (
    StillRunningLogger,
)
from exasol_integration_test_docker_environment.lib.docker.images.image_info import (
    ImageInfo,
    ImageState,
)

DOCKER_HUB_REGISTRY_URL_REGEX = r"^.*docker.io/"


# TODO align and extract save_path of DockerSaveImageTask and load_path of DockerLoadImageTask
class DockerSaveImageBaseTask(DockerBaseTask):
    image_name: str = luigi.Parameter()
    force_save: bool = luigi.BoolParameter(
        default=False, visibility=luigi.parameter.ParameterVisibility.HIDDEN
    )
    save_path: str = luigi.Parameter(
        visibility=luigi.parameter.ParameterVisibility.HIDDEN
    )

    def register_required(self):
        task = self.get_docker_image_task()
        self._image_info_future = self.register_dependency(task)

    def get_docker_image_task(self):
        raise AbstractMethodException()

    def run_task(self) -> None:
        image_info: ImageInfo = self.get_values_from_future(self._image_info_future)
        tag_for_save: str = self.get_tag_for_save(image_info)
        save_file_path = pathlib.Path(
            self.save_path, f"{image_info.get_target_complete_name()}.tar"
        )
        was_build = image_info.image_state == ImageState.WAS_BUILD.name
        if was_build or self.force_save or not save_file_path.exists():
            self.save_image(image_info, tag_for_save, save_file_path)
        self.return_object(image_info)

    def get_tag_for_save(self, image_info: ImageInfo):
        tag_for_save = re.sub(
            DOCKER_HUB_REGISTRY_URL_REGEX, "", image_info.get_target_complete_name()
        )
        return tag_for_save

    def save_image(
        self, image_info: ImageInfo, tag_for_save: str, save_file_path: pathlib.Path
    ):
        self.remove_save_file_if_necassary(save_file_path)
        with self._get_docker_client() as docker_client:
            image = docker_client.images.get(image_info.get_target_complete_name())
            generator = image.save(named=tag_for_save)
            self.write_image_to_file(save_file_path, image_info, generator)

    def remove_save_file_if_necassary(self, save_file_path: pathlib.Path):
        save_file_path.parent.mkdir(exist_ok=True, parents=True)
        if save_file_path.exists():
            os.remove(str(save_file_path))

    def write_image_to_file(
        self,
        save_file_path: pathlib.Path,
        image_info: ImageInfo,
        output_generator: Generator,
    ):
        self.logger.info(
            f"Saving image {image_info.get_target_complete_name()} to file {save_file_path}"
        )
        with save_file_path.open("wb") as file:
            still_running_logger = StillRunningLogger(
                self.logger, "save image %s" % image_info.get_target_complete_name()
            )
            for chunk in output_generator:
                still_running_logger.log()
                file.write(chunk)



import luigi

from exasol_integration_test_docker_environment.lib.base.docker_base_task import (
    DockerBaseTask,
)
from exasol_integration_test_docker_environment.lib.base.json_pickle_parameter import (
    JsonPickleParameter,
)
from exasol_integration_test_docker_environment.lib.models.data.database_credentials import (
    DatabaseCredentialsParameter,
)
from exasol_integration_test_docker_environment.lib.models.data.database_info import (
    DatabaseInfo,
)


class WaitForTestExternalDatabase(DockerBaseTask, DatabaseCredentialsParameter):
    environment_name: str = luigi.Parameter()
    database_info: DatabaseInfo = JsonPickleParameter(DatabaseInfo, significant=False)  # type: ignore
    db_startup_timeout_in_seconds: int = luigi.IntParameter(
        default=1 * 60, significant=False
    )
    attempt: int = luigi.IntParameter(default=1)

    def run_task(self) -> None:
        # Since we can't assume that the test container exists, we cannot connect easily here
        # to the external database (correct way would be by using an SQL client).
        # For now, we simply assume that the external database is already ready and return just True.
        self.return_object(True)



from pathlib import (
    Path,
    PurePath,
)

import luigi

from exasol_integration_test_docker_environment.abstract_method_exception import (
    AbstractMethodException,
)
from exasol_integration_test_docker_environment.lib.base.docker_base_task import (
    DockerBaseTask,
)
from exasol_integration_test_docker_environment.lib.base.json_pickle_parameter import (
    JsonPickleParameter,
)
from exasol_integration_test_docker_environment.lib.models.data.database_credentials import (
    DatabaseCredentialsParameter,
)
from exasol_integration_test_docker_environment.lib.models.data.environment_info import (
    EnvironmentInfo,
)


class PopulateTestDataToDatabase(DockerBaseTask, DatabaseCredentialsParameter):

    environment_name: str = luigi.Parameter()
    test_environment_info: EnvironmentInfo = JsonPickleParameter(
        EnvironmentInfo, significant=False
    )  # type: ignore

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._test_container_info = self.test_environment_info.test_container_info
        self._database_info = self.test_environment_info.database_info

    def run_task(self) -> None:
        self.logger.warning("Uploading data")
        username = self.db_user
        password = self.db_password
        data_path_within_test_container = self.get_data_path_within_test_container()
        data_file_within_data_path = self.get_data_file_within_data_path()
        with self._get_docker_client() as docker_client:
            assert self._test_container_info is not None
            test_container = docker_client.containers.get(
                self._test_container_info.container_name
            )
            cmd = (
                f"cd {data_path_within_test_container}; "
                f"$EXAPLUS -c '{self._database_info.host}:{self._database_info.ports.database}' "
                f"-x -u '{username}' -p '{password}' -f {data_file_within_data_path} "
                f"-jdbcparam 'validateservercertificate=0'"
            )

            bash_cmd = f"""bash -c "{cmd}" """
            exit_code, output = test_container.exec_run(cmd=bash_cmd)
        self.write_logs(output.decode("utf-8"))
        if exit_code != 0:
            raise Exception(
                "Failed to populate the database with data.\nLog: %s" % cmd
                + "\n"
                + output.decode("utf-8")
            )

    def write_logs(self, output: str):
        log_file = Path(self.get_log_path(), "log")
        with log_file.open("w") as file:
            file.write(output)

    def get_data_path_within_test_container(self) -> PurePath:
        raise AbstractMethodException()

    def get_data_file_within_data_path(self) -> PurePath:
        raise AbstractMethodException()



import luigi

from exasol_integration_test_docker_environment.lib.base.docker_base_task import (
    DockerBaseTask,
)
from exasol_integration_test_docker_environment.lib.base.json_pickle_parameter import (
    JsonPickleParameter,
)
from exasol_integration_test_docker_environment.lib.docker.images.image_info import (
    ImageInfo,
)


class DockerImageCreatorBaseTask(DockerBaseTask):
    image_name: str = luigi.Parameter()
    # ParameterVisibility needs to be hidden instead of private, because otherwise a MissingParameter gets thrown
    image_info: ImageInfo = JsonPickleParameter(
        ImageInfo,
        visibility=luigi.parameter.ParameterVisibility.HIDDEN,
        significant=True,
    )  # type: ignore



import getpass
import os
from typing import (
    Optional,
)

import luigi
from luigi.parameter import ParameterVisibility

from exasol_integration_test_docker_environment.cli.options.docker_repository_options import (
    DEFAULT_DOCKER_REPOSITORY_NAME,
)


class source_docker_repository_config(luigi.Config):
    repository_name: str = luigi.Parameter(default=DEFAULT_DOCKER_REPOSITORY_NAME)
    tag_prefix: str = luigi.Parameter(default="")
    username: Optional[str] = luigi.OptionalParameter(
        default=None, significant=False, visibility=ParameterVisibility.PRIVATE
    )
    password: Optional[str] = luigi.OptionalParameter(
        default=None, significant=False, visibility=ParameterVisibility.PRIVATE
    )


class target_docker_repository_config(luigi.Config):
    repository_name: str = luigi.Parameter(default=DEFAULT_DOCKER_REPOSITORY_NAME)
    tag_prefix: str = luigi.Parameter(default="")
    username: Optional[str] = luigi.OptionalParameter(
        default=None, significant=False, visibility=ParameterVisibility.PRIVATE
    )
    password: Optional[str] = luigi.OptionalParameter(
        default=None, significant=False, visibility=ParameterVisibility.PRIVATE
    )


class docker_build_arguments(luigi.Config):
    transparent: dict[str, str] = luigi.DictParameter(default={})
    image_changing: dict[str, str] = luigi.DictParameter(default={})
    secret: dict[str, str] = luigi.DictParameter(
        default={},
        description="Will not be saved somewhere, but are also assumed to be transparent",
        visibility=ParameterVisibility.PRIVATE,
    )


def set_docker_repository_config(
    docker_password: Optional[str],
    docker_repository_name: Optional[str],
    docker_username: Optional[str],
    tag_prefix: str,
    kind: str,
):
    config_class = f"{kind}_docker_repository_config"
    luigi.configuration.get_config().set(config_class, "tag_prefix", tag_prefix)
    if docker_repository_name is not None:
        luigi.configuration.get_config().set(
            config_class, "repository_name", docker_repository_name
        )
    password_environment_variable_name = f"{kind.upper()}_DOCKER_PASSWORD"
    if docker_username is not None:
        luigi.configuration.get_config().set(config_class, "username", docker_username)
        if docker_password is not None:
            luigi.configuration.get_config().set(
                config_class, "password", docker_password
            )
        elif password_environment_variable_name in os.environ:
            print(
                f"Using password from environment variable {password_environment_variable_name}"
            )
            password = os.environ[password_environment_variable_name]
            luigi.configuration.get_config().set(config_class, "password", password)
        else:
            password = getpass.getpass(
                f"{kind.capitalize()} Docker Registry Password for User %s:"
                % docker_username
            )
            luigi.configuration.get_config().set(config_class, "password", password)



import copy
import importlib
from collections.abc import (
    Generator,
    Iterator,
)
from typing import (
    Union,
)

import luigi

from exasol_integration_test_docker_environment.lib.base.docker_base_task import (
    DockerBaseTask,
)
from exasol_integration_test_docker_environment.lib.base.json_pickle_parameter import (
    JsonPickleParameter,
)
from exasol_integration_test_docker_environment.lib.docker.images.create.docker_image_build_task import (
    DockerBuildImageTask,
)
from exasol_integration_test_docker_environment.lib.docker.images.create.docker_image_load_task import (
    DockerLoadImageTask,
)
from exasol_integration_test_docker_environment.lib.docker.images.create.docker_image_pull_task import (
    DockerPullImageTask,
)
from exasol_integration_test_docker_environment.lib.docker.images.image_info import (
    ImageInfo,
    ImageState,
)
from exasol_integration_test_docker_environment.lib.docker.images.required_task_info import (
    RequiredTaskInfo,
    RequiredTaskInfoDict,
)


class DockerCreateImageTask(DockerBaseTask):
    image_name: str = luigi.Parameter()
    # ParameterVisibility needs to be hidden instead of private, because otherwise a MissingParameter gets thrown
    image_info: ImageInfo = JsonPickleParameter(
        ImageInfo,
        visibility=luigi.parameter.ParameterVisibility.HIDDEN,
        significant=True,
    )  # type: ignore

    def run_task(
        self,
    ) -> Iterator[
        Union[DockerBuildImageTask, DockerLoadImageTask, DockerPullImageTask]
    ]:
        new_image_info = yield from self.build(self.image_info)
        self.return_object(new_image_info)

    def build(self, image_info: ImageInfo) -> Generator[
        Union[DockerBuildImageTask, DockerLoadImageTask, DockerPullImageTask],
        None,
        ImageInfo,
    ]:
        if image_info.image_state == ImageState.NEEDS_TO_BE_BUILD.name:
            build_img_task: DockerBuildImageTask = self.create_child_task(
                DockerBuildImageTask, image_name=self.image_name, image_info=image_info
            )
            yield from self.run_dependencies(build_img_task)
            image_info.image_state = ImageState.WAS_BUILD.name  # TODO clone and change
            return image_info
        elif image_info.image_state == ImageState.CAN_BE_LOADED.name:
            load_img_task: DockerLoadImageTask = self.create_child_task(
                DockerLoadImageTask, image_name=self.image_name, image_info=image_info
            )
            yield from self.run_dependencies(load_img_task)
            image_info.image_state = ImageState.WAS_LOADED.name
            return image_info
        elif image_info.image_state == ImageState.REMOTE_AVAILABLE.name:
            pull_img_task: DockerPullImageTask = self.create_child_task(
                DockerPullImageTask, image_name=self.image_name, image_info=image_info
            )
            yield from self.run_dependencies(pull_img_task)
            image_info.image_state = ImageState.WAS_PULLED.name
            return image_info
        elif image_info.image_state == ImageState.TARGET_LOCALLY_AVAILABLE.name:
            image_info.image_state = ImageState.USED_LOCAL.name
            return image_info
        elif image_info.image_state == ImageState.SOURCE_LOCALLY_AVAILABLE.name:
            image_info.image_state = ImageState.WAS_TAGED.name
            self.rename_source_image_to_target_image(image_info)
            return image_info
        else:
            raise Exception(
                "Task %s: Image state %s not supported for image %s",
                self.task_id,
                image_info.image_state,
                image_info.get_target_complete_name(),
            )

    def rename_source_image_to_target_image(self, image_info) -> None:
        with self._get_docker_client() as docker_client:
            docker_client.images.get(image_info.get_source_complete_name()).tag(
                repository=image_info.target_repository_name,
                tag=image_info.get_target_complete_tag(),
            )


class DockerCreateImageTaskWithDeps(DockerCreateImageTask):
    # ParameterVisibility needs to be hidden instead of private, because otherwise a MissingParameter gets thrown
    required_task_infos: RequiredTaskInfoDict = JsonPickleParameter(
        RequiredTaskInfoDict,
        visibility=luigi.parameter.ParameterVisibility.HIDDEN,
        significant=True,
    )  # type: ignore

    def register_required(self) -> None:
        self.required_tasks = {
            key: self.create_required_task(required_task_info)
            for key, required_task_info in self.required_task_infos.infos.items()
        }
        self.futures = self.register_dependencies(self.required_tasks)

    def create_required_task(
        self, required_task_info: RequiredTaskInfo
    ) -> DockerCreateImageTask:
        module = importlib.import_module(required_task_info.module_name)
        class_ = getattr(module, required_task_info.class_name)
        instance = self.create_child_task(class_, **required_task_info.params)
        return instance

    def run_task(
        self,
    ) -> Iterator[
        Union[DockerBuildImageTask, DockerLoadImageTask, DockerPullImageTask]
    ]:
        image_infos = self.get_values_from_futures(self.futures)
        image_info = copy.copy(self.image_info)
        image_info.depends_on_images = image_infos
        new_image_info = yield from self.build(image_info)
        self.return_object(new_image_info)



from datetime import datetime
from enum import (
    Enum,
    auto,
)
from typing import (
    Any,
    Optional,
)

from exasol_integration_test_docker_environment.lib.base.info import Info


class ImageState(Enum):
    NOT_EXISTING = auto()
    # After analyze phase or if build phase did touch the image
    NEEDS_TO_BE_BUILD = auto()
    TARGET_LOCALLY_AVAILABLE = auto()
    SOURCE_LOCALLY_AVAILABLE = auto()
    REMOTE_AVAILABLE = auto()
    CAN_BE_LOADED = auto()
    # After build phase
    WAS_BUILD = auto()
    USED_LOCAL = auto()
    WAS_PULLED = auto()
    WAS_LOADED = auto()
    WAS_TAGED = auto()


class ImageDescription:
    def __init__(
        self,
        dockerfile: str,
        image_changing_build_arguments: dict[str, Any],
        transparent_build_arguments: dict[str, Any],
        mapping_of_build_files_and_directories: dict[str, str],
    ) -> None:
        self.transparent_build_arguments = transparent_build_arguments
        self.image_changing_build_arguments = image_changing_build_arguments
        self.mapping_of_build_files_and_directories = (
            mapping_of_build_files_and_directories
        )
        self.dockerfile = dockerfile

    def __repr__(self):
        return f"{self.__class__.__name__}({self.__dict__})"


class ImageInfo(Info):
    DOCKER_TAG_LENGTH_LIMIT = 128
    MAX_TAG_SURPLUS = 30

    def __init__(
        self,
        source_repository_name: str,
        target_repository_name: str,
        source_tag: str,
        target_tag: str,
        hash_value: str,
        commit: str,
        image_description: Optional[ImageDescription],
        build_name: str = "",
        build_date_time: datetime = datetime.utcnow(),
        image_state: Optional[ImageState] = ImageState.NOT_EXISTING,
        depends_on_images: Optional[dict[str, "ImageInfo"]] = None,
    ) -> None:
        self.build_name = build_name
        self.date_time = str(build_date_time)
        self.commit = commit
        self.target_repository_name = target_repository_name
        self.source_repository_name = source_repository_name
        self.image_description = image_description
        self.image_state: Optional[str] = None
        if isinstance(image_state, ImageState):
            self.image_state = image_state.name
        elif isinstance(image_state, str):
            self.image_state = ImageState[image_state].name
        elif image_state is not None:
            raise TypeError(f"{type(image_state)} for image_state not supported")
        self.depends_on_images = depends_on_images
        self.source_tag = source_tag
        self.target_tag = target_tag
        self.hash = hash_value
        self.check_complete_tag_length(self.source_tag)
        self.check_complete_tag_length(self.target_tag)

    def check_complete_tag_length(self, tag):
        complete_tag_length_limit = self.DOCKER_TAG_LENGTH_LIMIT + self.MAX_TAG_SURPLUS
        complete_tag = self._create_complete_tag(tag)
        if len(complete_tag) > complete_tag_length_limit:
            raise Exception(
                f"Complete Tag to long by {len(complete_tag) - complete_tag_length_limit}:  {complete_tag}"
            )

    def get_target_complete_name(self):
        return f"{self.target_repository_name}:{self.get_target_complete_tag()}"

    def get_source_complete_name(self):
        return f"{self.source_repository_name}:{self.get_source_complete_tag()}"

    def get_source_complete_tag(self):
        return self._create_truncated_complete_tag(self.source_tag)

    def get_target_complete_tag(self):
        return self._create_truncated_complete_tag(self.target_tag)

    def _create_truncated_complete_tag(self, tag: str) -> str:
        # we must truncate the tag to 128 characters, because this is the limit of docker tags
        # refer here https://docs.docker.com/engine/reference/commandline/tag/
        complete_tag = self._create_complete_tag(tag)
        truncated_tag = complete_tag[:128]
        return truncated_tag

    def _create_complete_tag(self, tag):
        if self.hash == "":
            return f"{tag}"
        else:
            return f"{tag}_{self.hash}"



import docker

from exasol_integration_test_docker_environment.lib.models.config.docker_config import (
    source_docker_repository_config,
    target_docker_repository_config,
)


class ContextDockerClient:
    def __init__(self, **kwargs) -> None:
        self.kwargs = kwargs
        self._client = None

    def __enter__(self):

        self._client = docker.from_env(**self.kwargs)
        self.login(self._client, source_docker_repository_config())
        self.login(self._client, target_docker_repository_config())
        return self._client

    def login(self, client: docker.client.DockerClient, docker_repository_config):
        if (
            docker_repository_config.username is not None
            and docker_repository_config.password is not None
        ):
            client.login(
                username=docker_repository_config.username,
                password=docker_repository_config.password,
            )

    def __exit__(self, type_, value, traceback):
        if self._client is not None:
            self._client.close()
        self._client = None



from exasol_integration_test_docker_environment.lib.base.dependency_logger_base_task import (
    DependencyLoggerBaseTask,
)
from exasol_integration_test_docker_environment.lib.base.docker_parameter import (
    DockerParameter,
)
from exasol_integration_test_docker_environment.lib.docker import ContextDockerClient


class DockerBaseTask(DependencyLoggerBaseTask, DockerParameter):

    def _get_docker_client(self) -> ContextDockerClient:
        return ContextDockerClient(timeout=self.timeout)



import base64
import os
import tempfile
from pathlib import Path
from string import Template
from typing import (
    Optional,
    Union,
)

import paramiko
import portalocker

_LOCK_FILE = "$TMP/$MODULE-ssh-access.lock"
_DEFAULT_CACHE_DIR = "$HOME/.cache/exasol/$MODULE"


def _path(template: str) -> Path:
    return Path(
        Template(template).substitute(
            TMP=tempfile.gettempdir(),
            HOME=os.path.expanduser("~"),
            MODULE="itde",
        )
    )


class SshKeyCache:
    def __init__(self, directory: Optional[Path] = None):
        self._directory = directory if directory else _path(_DEFAULT_CACHE_DIR)

    @property
    def directory(self) -> Path:
        return self._directory

    @property
    def private_key(self) -> Path:
        return self._directory / "id_rsa"

    @property
    def public_key(self) -> Path:
        return self._directory / "id_rsa.pub"


class SshKey:
    """This class hosts the private key for SSH access to a Docker Container
    and offers some convenience methods for writing the private or public key
    to a file or getting a string representation to write it to an
    authorized_keys file.

    The maker methods enable to either generate a new random key or to read it
    from a file, e.g. named "id_rsa".

    When creating a new instance based on a directory containing a file id_rsa
    then the maker method will read the private key from this file.

    If the directory does not contain this file then the maker method will
    generate a new random SSH key, will write the private key to this file and
    the public key to file id_rsa.pub in the same directory.

    If the directory does not exist, then the maker method will create the directory
    and continue with generating a new random key.

    ITDE uses python library portalocker to guarantee that the key files are
    accessed only by a single process at a time.
    """

    def __init__(self, private_key: paramiko.RSAKey) -> None:
        self.private = private_key

    def write_private_key(self, path: str) -> "SshKey":
        self.private.write_private_key_file(path)  # uses 0o600 = user r/w
        return self

    def public_key_as_string(self, comment="") -> str:
        b64 = base64.b64encode(self.private.asbytes()).decode("ascii")
        return f"ssh-rsa {b64} {comment}"

    def write_public_key(self, path: str, comment="") -> "SshKey":
        def opener(path, flags):
            return os.open(path, flags, 0o600)

        content = self.public_key_as_string(comment)
        # Windows does not support kwarg mode=0o600 here
        with open(path, "w", opener=opener) as file:
            print(content, file=file)
        return self

    @classmethod
    def read_from(cls, private_key_file: Union[Path, str]) -> "SshKey":
        with open(private_key_file) as file:
            rsa_key = paramiko.RSAKey.from_private_key(file)
        return SshKey(rsa_key)

    @classmethod
    def generate(cls) -> "SshKey":
        rsa_key = paramiko.RSAKey.generate(bits=4096)
        return SshKey(rsa_key)

    @classmethod
    def from_cache(cls, cache_directory: Optional[Path] = None) -> "SshKey":
        cache = SshKeyCache(cache_directory)
        priv = cache.private_key
        with portalocker.Lock(_path(_LOCK_FILE), "wb", timeout=10) as fh:
            if priv.exists():
                return cls.read_from(priv)
            # mode 0o700 = rwx permissions only for the current user
            # is required for the directory to enable to create files inside
            os.makedirs(cache.directory, mode=0o700, exist_ok=True)
            return (
                cls.generate()
                .write_private_key(str(priv))
                .write_public_key(str(cache.public_key))
            )



from exasol_integration_test_docker_environment.abstract_method_exception import (
    AbstractMethodException,
)


class BucketFSSyncChecker:
    def prepare_upload(self):
        raise AbstractMethodException()

    def wait_for_bucketfs_sync(self):
        raise AbstractMethodException()



from enum import Enum

import luigi


class WriteLogFilesToConsole(Enum):
    only_error = 1
    all = 2


class log_config(luigi.Config):
    write_log_files_to_console = luigi.EnumParameter(
        enum=WriteLogFilesToConsole, default=WriteLogFilesToConsole.only_error
    )
    log_task_is_still_running = luigi.BoolParameter(default=False)
    seconds_between_still_running_logs = luigi.IntParameter(default=60)



import json
from typing import (
    Optional,
)

import luigi

from exasol_integration_test_docker_environment.cli.options.system_options import (
    DEFAULT_OUTPUT_DIRECTORY,
)


class build_config(luigi.Config):
    force_pull: bool = luigi.BoolParameter(default=False)
    force_load: bool = luigi.BoolParameter(default=False)
    force_rebuild: bool = luigi.BoolParameter(default=False)
    force_rebuild_from: list[str] = luigi.ListParameter(default=[])
    log_build_context_content: bool = luigi.BoolParameter(default=False)
    # keep_build_context = luigi.BoolParameter(False)
    temporary_base_directory: Optional[str] = luigi.OptionalParameter(default=None)
    output_directory: str = luigi.Parameter(default=DEFAULT_OUTPUT_DIRECTORY)
    cache_directory: Optional[str] = luigi.OptionalParameter(default="")
    build_name: Optional[str] = luigi.OptionalParameter(default="")


def set_output_directory(output_directory: Optional[str]):
    if output_directory is not None:
        luigi.configuration.get_config().set(
            "build_config", "output_directory", output_directory
        )


def set_build_config(
    force_rebuild: bool,
    force_rebuild_from: tuple[str, ...],
    force_pull: bool,
    log_build_context_content: bool,
    output_directory: Optional[str],
    temporary_base_directory: Optional[str],
    cache_directory: Optional[str],
    build_name: Optional[str],
):
    luigi.configuration.get_config().set(
        "build_config", "force_rebuild", str(force_rebuild)
    )
    luigi.configuration.get_config().set(
        "build_config", "force_rebuild_from", json.dumps(force_rebuild_from)
    )
    luigi.configuration.get_config().set("build_config", "force_pull", str(force_pull))
    set_output_directory(output_directory)
    if temporary_base_directory is not None:
        luigi.configuration.get_config().set(
            "build_config", "temporary_base_directory", temporary_base_directory
        )
    if cache_directory is not None:
        luigi.configuration.get_config().set(
            "build_config", "cache_directory", cache_directory
        )
    if build_name is not None:
        luigi.configuration.get_config().set("build_config", "build_name", build_name)
    luigi.configuration.get_config().set(
        "build_config", "log_build_context_content", str(log_build_context_content)
    )



import logging

from docker.models.containers import Container

from exasol_integration_test_docker_environment.lib.docker import ContextDockerClient


def remove_docker_container(containers: list[str]):
    """
    Removes the given container using docker API.
    """
    with ContextDockerClient() as docker_client:
        for container in containers:
            try:
                docker_client.containers.get(container).remove(force=True)
            except Exception as e:
                logging.error(e)


def default_bridge_ip_address(container: Container) -> str:
    container.reload()
    return container.attrs["NetworkSettings"]["Networks"]["bridge"]["IPAddress"]



from typing import Optional

from exasol_integration_test_docker_environment.lib.models.data.database_info import (
    DatabaseInfo,
)
from exasol_integration_test_docker_environment.lib.models.data.docker_network_info import (
    DockerNetworkInfo,
)
from exasol_integration_test_docker_environment.lib.models.data.docker_volume_info import (
    DockerVolumeInfo,
)
from exasol_integration_test_docker_environment.lib.models.data.environment_type import (
    EnvironmentType,
)
from exasol_integration_test_docker_environment.lib.test_environment.abstract_spawn_test_environment import (
    AbstractSpawnTestEnvironment,
)
from exasol_integration_test_docker_environment.lib.test_environment.database_waiters.wait_for_external_database import (
    WaitForTestExternalDatabase,
)
from exasol_integration_test_docker_environment.lib.test_environment.parameter.external_test_environment_parameter import (
    ExternalDatabaseHostParameter,
    ExternalDatabaseXMLRPCParameter,
)
from exasol_integration_test_docker_environment.lib.test_environment.prepare_network_for_test_environment import (
    PrepareDockerNetworkForTestEnvironment,
)
from exasol_integration_test_docker_environment.lib.test_environment.setup_external_database_host import (
    SetupExternalDatabaseHost,
)


class SpawnTestEnvironmentWithExternalDB(
    AbstractSpawnTestEnvironment,
    ExternalDatabaseXMLRPCParameter,
    ExternalDatabaseHostParameter,
):

    def get_environment_type(self):
        return EnvironmentType.external_db

    def create_network_task(self, attempt: int):
        return self.create_child_task_with_common_params(
            PrepareDockerNetworkForTestEnvironment,
            reuse=self.reuse_test_container,
            attempt=attempt,
            test_container_name=self.test_container_name,
            network_name=self.network_name,
            no_cleanup_after_success=self.reuse_test_container,
            no_cleanup_after_failure=self.reuse_test_container,
        )

    def create_spawn_database_task(
        self,
        network_info: DockerNetworkInfo,
        certificate_volume_info: Optional[DockerVolumeInfo],
        attempt: int,
    ):
        if certificate_volume_info is not None:
            raise ValueError(
                "Certificate volume must be None when using external database"
            )

        return self.create_child_task_with_common_params(
            SetupExternalDatabaseHost, network_info=network_info, attempt=attempt
        )

    def create_wait_for_database_task(
        self, attempt: int, database_info: DatabaseInfo
    ) -> WaitForTestExternalDatabase:
        return self.create_child_task_with_common_params(
            WaitForTestExternalDatabase, database_info=database_info, attempt=attempt
        )



from pathlib import Path
from typing import (
    Optional,
    Union,
)

import importlib_resources
import luigi
import netaddr
from docker.models.containers import Container
from docker.transport import unixconn

from exasol_integration_test_docker_environment.lib import PACKAGE_NAME
from exasol_integration_test_docker_environment.lib.base.docker_base_task import (
    DockerBaseTask,
)
from exasol_integration_test_docker_environment.lib.base.json_pickle_parameter import (
    JsonPickleParameter,
)
from exasol_integration_test_docker_environment.lib.docker.images.image_info import (
    ImageInfo,
    ImageState,
)
from exasol_integration_test_docker_environment.lib.models.data.container_info import (
    ContainerInfo,
)
from exasol_integration_test_docker_environment.lib.models.data.docker_network_info import (
    DockerNetworkInfo,
)
from exasol_integration_test_docker_environment.lib.test_environment.analyze_test_container import (
    DockerTestContainerBuild,
)
from exasol_integration_test_docker_environment.lib.test_environment.docker_container_copy import (
    copy_script_to_container,
)
from exasol_integration_test_docker_environment.lib.test_environment.parameter.test_container_parameter import (
    TestContainerParameter,
)


class SpawnTestContainer(DockerBaseTask, TestContainerParameter):
    environment_name: str = luigi.Parameter()
    test_container_name: str = luigi.Parameter()
    network_info: DockerNetworkInfo = JsonPickleParameter(
        DockerNetworkInfo, significant=False
    )  # type: ignore
    ip_address_index_in_subnet: int = luigi.IntParameter(significant=False)
    attempt: int = luigi.IntParameter(default=1)
    reuse_test_container: bool = luigi.BoolParameter(default=False, significant=False)
    no_test_container_cleanup_after_success: bool = luigi.BoolParameter(
        default=False, significant=False
    )
    no_test_container_cleanup_after_failure: bool = luigi.BoolParameter(
        default=False, significant=False
    )
    docker_runtime: Optional[str] = luigi.OptionalParameter(
        default=None, significant=False
    )
    certificate_volume_name: Optional[str] = luigi.OptionalParameter(
        default=None, significant=False
    )

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        if self.ip_address_index_in_subnet < 0:
            raise Exception(
                "ip_address_index_in_subnet needs to be greater than 0 got %s"
                % self.ip_address_index_in_subnet
            )

    def register_required(self) -> None:
        self.test_container_image_future = self.register_dependency(
            self.create_child_task(
                task_class=DockerTestContainerBuild,
                test_container_content=self.test_container_content,
            )
        )

    def is_reuse_possible(self) -> bool:
        test_container_image_info: ImageInfo = self.get_values_from_futures(
            self.test_container_image_future
        )[
            "test-container"
        ]  # type: ignore
        test_container = None
        with self._get_docker_client() as docker_client:
            try:
                test_container = docker_client.containers.get(self.test_container_name)
            except Exception as e:
                pass
            ret_val: bool = bool(
                self.network_info.reused
                and self.reuse_test_container
                and test_container is not None
                and test_container_image_info.get_target_complete_name()
                in test_container.image.tags
                and test_container_image_info.image_state == ImageState.USED_LOCAL.name
            )

        return ret_val

    def run_task(self) -> None:
        subnet = netaddr.IPNetwork(self.network_info.subnet)
        ip_address = str(subnet[2 + self.ip_address_index_in_subnet])
        container_info = None

        if self.is_reuse_possible():
            container_info = self._try_to_reuse_test_container(
                ip_address, self.network_info
            )
        if container_info is None:
            container_info = self._create_test_container(ip_address, self.network_info)
        with self._get_docker_client() as docker_client:
            docker_client.containers.get(self.test_container_name)
        self._copy_runtime_targets()
        self.return_object(container_info)

    def _copy_runtime_targets(self) -> None:
        self.logger.info(
            "Copy runtime targets in test container %s.", self.test_container_name
        )
        with self._get_docker_client() as docker_client:
            test_container = docker_client.containers.get(self.test_container_name)
            for runtime_mapping in self.test_container_content.runtime_mappings:
                if runtime_mapping.deployment_target is not None:
                    self.logger.warning(
                        f"Copy runtime target {runtime_mapping.target} "
                        f"in test container {self.test_container_name}."
                    )
                    try:
                        test_container.exec_run(
                            cmd=f"rm -r {runtime_mapping.deployment_target}"
                        )
                    except:
                        pass
                    test_container.exec_run(
                        cmd=f"cp -r {runtime_mapping.target} {runtime_mapping.deployment_target}"
                    )

    def _try_to_reuse_test_container(
        self, ip_address: str, network_info: DockerNetworkInfo
    ) -> Optional[ContainerInfo]:
        self.logger.info("Try to reuse test container %s", self.test_container_name)
        container_info = None
        try:
            network_aliases = self._get_network_aliases()
            container_info = self.create_container_info(
                ip_address, network_aliases, network_info
            )
        except Exception as e:
            self.logger.warning(
                "Tried to reuse test container %s, but got Exeception %s. "
                "Fallback to create new database.",
                self.test_container_name,
                e,
            )
        return container_info

    def _create_test_container(
        self, ip_address, network_info: DockerNetworkInfo
    ) -> ContainerInfo:
        self._remove_container(self.test_container_name)
        self.logger.info(f"Creating new test container {self.test_container_name}")
        test_container_image_info = self.get_values_from_futures(
            self.test_container_image_future
        )["test-container"]

        volumes: dict[Union[str, Path], dict[str, str]] = {}
        for runtime_mapping in self.test_container_content.runtime_mappings:
            volumes[runtime_mapping.source.absolute()] = {
                "bind": runtime_mapping.target,
                "mode": "rw",
            }
        if self.certificate_volume_name is not None:
            volumes[self.certificate_volume_name] = {
                "bind": "/certificates",
                "mode": "ro",
            }

        with self._get_docker_client() as docker_client:
            docker_unix_sockets = [
                i
                for i in docker_client.api.adapters.values()
                if isinstance(i, unixconn.UnixHTTPAdapter)
            ]
            if len(docker_unix_sockets) > 0:
                host_docker_socker_path = docker_unix_sockets[0].socket_path
                volumes[host_docker_socker_path] = {
                    "bind": "/var/run/docker.sock",
                    "mode": "rw",
                }
            test_container = docker_client.containers.create(
                image=test_container_image_info.get_target_complete_name(),
                name=self.test_container_name,
                network_mode=None,
                command="sleep infinity",
                detach=True,
                volumes=volumes,
                labels={
                    "test_environment_name": self.environment_name,
                    "container_type": "test_container",
                },
                runtime=self.docker_runtime,
            )
            docker_network = docker_client.networks.get(network_info.network_name)
            network_aliases = self._get_network_aliases()
            docker_network.connect(
                test_container, ipv4_address=ip_address, aliases=network_aliases
            )
            test_container.start()
            self.register_certificates(test_container)
            container_info = self.create_container_info(
                ip_address, network_aliases, network_info
            )
            return container_info

    def _get_network_aliases(self) -> list[str]:
        network_aliases = ["test_container", self.test_container_name]
        return network_aliases

    def create_container_info(
        self,
        ip_address: str,
        network_aliases: list[str],
        network_info: DockerNetworkInfo,
    ) -> ContainerInfo:
        with self._get_docker_client() as docker_client:
            test_container = docker_client.containers.get(self.test_container_name)
            if test_container.status != "running":
                raise Exception(f"Container {self.test_container_name} not running")
            container_info = ContainerInfo(
                container_name=self.test_container_name,
                ip_address=ip_address,
                network_aliases=network_aliases,
                network_info=network_info,
            )
        return container_info

    def _get_export_directory(self):
        return self.get_values_from_future(self.export_directory_future)  # type: ignore

    def _remove_container(self, container_name: str):
        try:
            with self._get_docker_client() as docker_client:
                container = docker_client.containers.get(container_name)
                container.remove(force=True)
                self.logger.info(
                    f"Removed container: name: '{container_name}', id: '{container.short_id}'"
                )
        except Exception as e:
            pass

    def register_certificates(self, test_container: Container) -> None:
        if self.certificate_volume_name is not None:
            script_name = "install_root_certificate.sh"
            script = (
                importlib_resources.files(PACKAGE_NAME)
                / "certificate_resources"
                / script_name
            )
            script_location_in_container = f"scripts/{script_name}"
            copy_script_to_container(
                script.read_text(),
                script_location_in_container,
                test_container,
            )

            exit_code, output = test_container.exec_run(
                f"bash {script_location_in_container}"
            )
            if exit_code != 0:
                raise RuntimeError(
                    f"Error installing certificates:{output.decode('utf-8')}"
                )

    def cleanup_task(self, success: bool) -> None:
        if (success and not self.no_test_container_cleanup_after_success) or (
            not success and not self.no_test_container_cleanup_after_failure
        ):
            try:
                self.logger.info(f"Cleaning up container %s", self.test_container_name)
                self._remove_container(self.test_container_name)
            except Exception as e:
                self.logger.error(
                    f"Error during removing container %s: %s",
                    self.test_container_name,
                    e,
                )



import io
import tarfile
import time
from typing import Optional

from docker.models.containers import Container


class DockerContainerCopy:

    def __init__(self, container: Container) -> None:
        super().__init__()
        self.open = True
        self.file_like_object: Optional[io.BytesIO] = io.BytesIO()
        self.tar: Optional[tarfile.TarFile] = tarfile.open(
            fileobj=self.file_like_object, mode="x"
        )
        self.container = container

    def __del__(self):
        if self.open:
            self.tar.close()
            self.open = False
            self.tar = None
            self.file_like_object = None

    def is_open_or_raise(self):
        if not self.open:
            raise Exception("DockerContainerCopy not open")

    def add_string_to_file(self, name: str, string: str):
        self.is_open_or_raise()
        encoded = string.encode("utf-8")
        bytes_io = io.BytesIO(encoded)
        tar_info = tarfile.TarInfo(name=name)
        tar_info.mtime = time.time()
        tar_info.size = len(encoded)
        assert self.tar
        self.tar.addfile(tarinfo=tar_info, fileobj=bytes_io)

    def add_file(self, host_path: str, path_in_tar: str):
        self.is_open_or_raise()
        assert self.tar
        self.tar.add(host_path, path_in_tar)

    def copy(self, path_in_container: str):
        self.is_open_or_raise()
        assert self.tar
        self.tar.close()
        self.open = False
        self.tar = None
        assert self.file_like_object
        self.container.put_archive(
            path_in_container, self.file_like_object.getbuffer().tobytes()
        )
        self.file_like_object = None


def copy_script_to_container(script: str, script_path: str, container: Container):
    """
    Copy a script, given as string, to the container under specified location, relative to "/"
    """
    docker_container_copy = DockerContainerCopy(container)
    docker_container_copy.add_string_to_file(script_path, script)
    docker_container_copy.copy("/")



class SshInfo:
    """
    key_file contains path to the file containing the private key for SSH access.
    """

    def __init__(self, user: str, key_file: str) -> None:
        self.user = user
        self.key_file = key_file



import luigi
from luigi import Config


class DatabaseCredentials:
    def __init__(
        self, db_user: str, db_password: str, bucketfs_write_password: str
    ) -> None:
        self.bucketfs_write_password = bucketfs_write_password
        self.db_password = db_password
        self.db_user = db_user


class DatabaseCredentialsParameter(Config):
    db_user: str = luigi.Parameter()
    db_password: str = luigi.Parameter(
        significant=False, visibility=luigi.parameter.ParameterVisibility.HIDDEN
    )
    bucketfs_write_password: str = luigi.Parameter(
        significant=False, visibility=luigi.parameter.ParameterVisibility.HIDDEN
    )

    def get_database_credentials(self):
        return DatabaseCredentials(
            self.db_user, self.db_password, self.bucketfs_write_password
        )



from typing import (
    Optional,
    Union,
)

from exasol_integration_test_docker_environment.lib.base.info import Info
from exasol_integration_test_docker_environment.lib.models.data.container_info import (
    ContainerInfo,
)
from exasol_integration_test_docker_environment.lib.models.data.database_info import (
    DatabaseInfo,
)
from exasol_integration_test_docker_environment.lib.models.data.docker_network_info import (
    DockerNetworkInfo,
)
from exasol_integration_test_docker_environment.lib.models.data.environment_type import (
    EnvironmentType,
)


class EnvironmentInfo(Info):

    def __init__(
        self,
        name: str,
        env_type: Union[str, EnvironmentType],
        database_info: DatabaseInfo,
        test_container_info: Optional[ContainerInfo],
        network_info: DockerNetworkInfo,
    ) -> None:
        self.name = name
        self.type = env_type
        self.test_container_info = test_container_info
        self.database_info = database_info
        self.network_info = network_info



"""
The doctor module provides functionality to check the health of the `exasol_integration_test_docker_environment`
package and also provide help to find potential fixes.
"""

import sys
from collections.abc import (
    Callable,
    Iterable,
)
from enum import Enum

import docker
from docker.errors import DockerException
from exasol.error import ExaError

SUPPORTED_PLATFORMS = ["linux", "darwin"]


class HealthProblem(Enum):
    Unknown = ExaError(
        "E-ITDE-0",
        "Unknown issue.",
        ["An unknown error occurred, please contact the maintainer."],
        {},
    )

    UnixSocketNotAvailable = ExaError(
        "E-ITDE-1",
        "Could not find unix socket to connect to.",
        ["Make sure environment variable DOCKER_HOST is configured correctly."],
        {},
    )

    TargetPlatformNotSupported = ExaError(
        "E-ITDE-2",
        "The platform ITDE is running on is not supported.",
        ["Make sure you are using one of the following platforms: [linux, darwin]."],
        {},
    )


def diagnose_docker_daemon_not_available() -> Iterable[HealthProblem]:
    """Diagnose reasons why docker deamon is not available"""

    def _is_unix_socket_issue(message: str) -> bool:
        return "FileNotFoundError(2, 'No such file or directory')" in message

    errors = []
    try:
        _docker = docker.from_env()
    except DockerException as ex:
        msg = f"{ex}"
        if _is_unix_socket_issue(msg):
            errors.append(HealthProblem.UnixSocketNotAvailable)
        if len(errors) == 0:
            errors.append(HealthProblem.Unknown)
    return errors


def is_docker_daemon_available() -> bool:
    """
    Checks weather or not the docker daemon is available.
    """
    try:
        _docker = docker.from_env()
    except DockerException:
        return False
    return True


def is_supported_platform() -> bool:
    """
    Checks weather or not the current platform is supported.
    """
    return sys.platform in SUPPORTED_PLATFORMS


def health_checkup() -> Iterable[HealthProblem]:
    """
    Runs all known examinations

    return an iterator of error codes specifying which problems have been identified.
    """
    check_function = Callable[[], bool]
    diagnosis_function = Callable[[], Iterable[HealthProblem]]
    examinations: list[tuple[check_function, diagnosis_function]] = [
        (is_docker_daemon_available, diagnose_docker_daemon_not_available),
        (is_supported_platform, lambda: [HealthProblem.TargetPlatformNotSupported]),
    ]
    for is_fine, diagnosis in examinations:
        if not is_fine():
            yield from diagnosis()



import logging
import socket
import warnings
from collections.abc import Generator
from contextlib import ExitStack
from typing import (
    Optional,
)


def find_free_ports(num_ports: int) -> list[int]:
    def new_socket():
        return socket.socket(socket.AF_INET, socket.SOCK_STREAM)

    def bind(sock: socket.socket, port: int):
        sock.bind(("", port))
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

    def acquire_port_numbers(num_ports: int) -> Generator[int, None, None]:
        with ExitStack() as stack:
            sockets = [stack.enter_context(new_socket()) for dummy in range(num_ports)]
            for sock in sockets:
                bind(sock, 0)
                yield sock.getsockname()[1]

    def check_port_numbers(ports):
        with ExitStack() as stack:
            sockets_and_ports = [
                (stack.enter_context(new_socket()), port) for port in ports
            ]
            for sock, port in sockets_and_ports:
                bind(sock, port)

    ports = list(acquire_port_numbers(num_ports))
    check_port_numbers(ports)
    return ports


class PortsType(type):
    """
    The following properties are read-only class attributes:
    - default_ports
    - external
    - forward
    """

    @property
    def default_ports(self) -> "Ports":
        return Ports(database=8563, bucketfs=2580, ssh=22, bucketfs_https=2581)

    @property
    def external(self) -> "Ports":
        # For external databases SSH port might depend on version database.
        return Ports(database=8563, bucketfs=2580, ssh=None, bucketfs_https=2581)

    @property
    def forward(self) -> "Ports":
        return Ports(database=8563, bucketfs=2580, ssh=20002, bucketfs_https=2581)


class Ports(metaclass=PortsType):
    def __init__(
        self,
        database: Optional[int],
        bucketfs: Optional[int],
        ssh: Optional[int] = None,
        bucketfs_https: Optional[int] = None,
    ) -> None:
        if database is None:
            self._database: int = Ports.default_ports.database
        else:
            self._database = database
        if bucketfs is None:
            self._bucketfs_http: int = Ports.default_ports.bucketfs_http
        else:
            self._bucketfs_http = bucketfs
        self._ssh = ssh
        if bucketfs_https is None:
            self._bucketfs_https: int = Ports.default_ports.bucketfs_https
        else:
            self._bucketfs_https = bucketfs_https

    @property
    def bucketfs(self) -> int:
        warnings.warn(
            f"Property Ports.bucketfs is deprecated, use Ports.bucketfs_http instead.",
            DeprecationWarning,
        )
        return self._bucketfs_http

    @bucketfs.setter
    def bucketfs(self, value: int) -> None:
        warnings.warn(
            f"Property Ports.bucketfs is deprecated, use Ports.bucketfs_http instead.",
            DeprecationWarning,
        )
        self._bucketfs_http = value

    @property
    def bucketfs_http(self) -> int:
        return self._bucketfs_http

    @bucketfs_http.setter
    def bucketfs_http(self, value: int) -> None:
        warnings.warn(
            f"Setting values of class Ports after instantiation is deprecated.",
            DeprecationWarning,
        )
        self._bucketfs_http = value

    @property
    def bucketfs_https(self) -> int:
        return self._bucketfs_https

    @bucketfs_https.setter
    def bucketfs_https(self, value: int) -> None:
        warnings.warn(
            f"Setting values of class Ports after instantiation is deprecated.",
            DeprecationWarning,
        )
        self._bucketfs_https = value

    @property
    def database(self) -> int:
        return self._database

    @database.setter
    def database(self, value: int) -> None:
        warnings.warn(
            f"Setting values of class Ports after instantiation is deprecated.",
            DeprecationWarning,
        )
        self._database = value

    @property
    def ssh(self) -> Optional[int]:
        return self._ssh

    @ssh.setter
    def ssh(self, value: Optional[int]) -> None:
        warnings.warn(
            f"Setting values of class Ports after instantiation is deprecated.",
            DeprecationWarning,
        )
        self._ssh = value

    @classmethod
    def random_free(cls, ssh: bool = True) -> "Ports":
        count = 4 if ssh else 3
        ports = find_free_ports(count)
        return (
            Ports(ports[0], ports[1], None, ports[2])
            if not ssh
            else Ports(ports[0], ports[1], ports[2], ports[3])
        )



from typing import (
    Callable,
)


def find_images_by_tag(docker_client, condition: Callable[[str], bool]) -> list:
    images = docker_client.images.list()
    filter_images = [
        image
        for image in images
        if image.tags is not None
        and len(image.tags) > 0
        and any([condition(tag) for tag in image.tags])
    ]
    return filter_images



from typing import (
    Optional,
)

from exasol_integration_test_docker_environment.cli.options.docker_repository_options import (
    DEFAULT_DOCKER_REPOSITORY_NAME,
)
from exasol_integration_test_docker_environment.cli.options.system_options import (
    DEFAULT_OUTPUT_DIRECTORY,
)
from exasol_integration_test_docker_environment.lib.base.run_task import (
    generate_root_task,
    run_task,
)
from exasol_integration_test_docker_environment.lib.docker.images.image_info import (
    ImageInfo,
)
from exasol_integration_test_docker_environment.lib.models.config.build_config import (
    set_build_config,
)
from exasol_integration_test_docker_environment.lib.models.config.docker_config import (
    set_docker_repository_config,
)
from exasol_integration_test_docker_environment.lib.models.data.test_container_content_description import (
    TestContainerContentDescription,
)
from exasol_integration_test_docker_environment.lib.test_environment.analyze_test_container import (
    AnalyzeTestContainer,
    DockerTestContainerBuild,
)
from exasol_integration_test_docker_environment.lib.utils.api_function_decorators import (
    no_cli_function,
)


@no_cli_function
def build_test_container(
    test_container_content: TestContainerContentDescription,
    force_rebuild: bool = False,
    force_rebuild_from: tuple[str, ...] = (),
    force_pull: bool = False,
    output_directory: str = DEFAULT_OUTPUT_DIRECTORY,
    temporary_base_directory: str = "/tmp",
    log_build_context_content: bool = False,
    cache_directory: Optional[str] = None,
    build_name: Optional[str] = None,
    source_docker_repository_name: str = DEFAULT_DOCKER_REPOSITORY_NAME,
    source_docker_tag_prefix: str = "",
    source_docker_username: Optional[str] = None,
    source_docker_password: Optional[str] = None,
    target_docker_repository_name: str = DEFAULT_DOCKER_REPOSITORY_NAME,
    target_docker_tag_prefix: str = "",
    target_docker_username: Optional[str] = None,
    target_docker_password: Optional[str] = None,
    workers: int = 5,
    task_dependencies_dot_file: Optional[str] = None,
    log_level: Optional[str] = None,
    use_job_specific_log_file: bool = False,
) -> ImageInfo:
    """
    This function builds all stages of the test container for the test environment.
    If stages are cached in a docker registry, the function is going to pull them,
    instead of building them.
    It returns the image info of the test-container.
    """
    set_build_config(
        force_rebuild,
        force_rebuild_from,
        force_pull,
        log_build_context_content,
        output_directory,
        temporary_base_directory,
        cache_directory,
        build_name,
    )
    # Use AnalyzeTestContainer to ensure that all luigi processes got it loaded
    analyze_task = AnalyzeTestContainer.__class__.__name__

    set_docker_repository_config(
        source_docker_password,
        source_docker_repository_name,
        source_docker_username,
        source_docker_tag_prefix,
        "source",
    )
    set_docker_repository_config(
        target_docker_password,
        target_docker_repository_name,
        target_docker_username,
        target_docker_tag_prefix,
        "target",
    )
    task_creator = lambda: generate_root_task(
        task_class=DockerTestContainerBuild,
        test_container_content=test_container_content,
    )
    image_infos = run_task(
        task_creator,
        workers,
        task_dependencies_dot_file,
        log_level=log_level,
        use_job_specific_log_file=use_job_specific_log_file,
    )
    return image_infos["test-container"]



from typing import (
    Optional,
)

import click

from exasol_integration_test_docker_environment.cli.cli import cli
from exasol_integration_test_docker_environment.cli.options.docker_repository_options import (
    docker_repository_options,
)
from exasol_integration_test_docker_environment.cli.options.system_options import (
    luigi_logging_options,
    output_directory_option,
    system_options,
    tempory_base_directory_option,
)
from exasol_integration_test_docker_environment.cli.options.test_environment_options import (
    DEFAULT_DISK_SIZE,
    DEFAULT_MEM_SIZE,
    docker_db_options,
)
from exasol_integration_test_docker_environment.cli.termination_handler import (
    TerminationHandler,
)
from exasol_integration_test_docker_environment.lib import api
from exasol_integration_test_docker_environment.lib.models.api_errors import (
    ArgumentConstraintError,
)
from exasol_integration_test_docker_environment.lib.utils.cli_function_decorators import (
    add_options,
)


@cli.command()
@click.option(
    "--environment-name",
    type=str,
    required=True,
    help="Name of the docker environment. This name gets used as suffix for the container db_container_<name> and test_container_<name>",
)
@click.option(
    "--database-port-forward",
    type=int,
    default=None,
    show_default=True,
    help="Host port to which the database port gets forwarded",
)
@click.option(
    "--bucketfs-port-forward",
    type=int,
    default=None,
    show_default=True,
    help="Host port to which the BucketFS HTTP port gets forwarded. Deprecated: Use '--bucketfs-http-port-forward' instead.",
)
@click.option(
    "--ssh-port-forward",
    type=int,
    default=None,
    show_default=True,
    help="Host port to which the SSH port gets forwarded. If not specified then ITDE selects a random free port.",
)
@click.option(
    "--db-mem-size",
    type=str,
    default=DEFAULT_MEM_SIZE,
    show_default=True,
    help="The main memory used by the database. Format <number> <unit>, e.g. 1 GiB. The minimum size is 1 GB, below that the database will not start.",
)
@click.option(
    "--db-disk-size",
    type=str,
    default=DEFAULT_DISK_SIZE,
    show_default=True,
    help="The disk size available for the database. Format <number> <unit>, e.g. 1 GiB. The minimum size is 100 MiB. However, the setup creates volume files with at least 2 GB larger size, because the database needs at least so much more disk.",
)
@click.option(
    "--nameserver",
    type=str,
    default=[],
    multiple=True,
    help="Add a nameserver to the list of DNS nameservers which the docker-db should use for resolving domain names. You can repeat this option to add further nameservers.",
)
@click.option(
    "--docker-runtime",
    type=str,
    default=None,
    show_default=True,
    help="The docker runtime used to start all containers",
)
@add_options(docker_db_options)
@add_options(docker_repository_options)
@add_options([output_directory_option])
@add_options([tempory_base_directory_option])
@add_options(system_options)
@add_options(luigi_logging_options)
@click.option(
    "--bucketfs-http-port-forward",
    type=int,
    default=None,
    show_default=True,
    help="Host port to which the BucketFS HTTP port gets forwarded",
)
@click.option(
    "--bucketfs-https-port-forward",
    type=int,
    default=None,
    show_default=True,
    help="Host port to which the BucketFS HTTPS port gets forwarded",
)
def spawn_test_environment(
    environment_name: str,
    database_port_forward: Optional[int],
    bucketfs_port_forward: Optional[int],
    ssh_port_forward: Optional[int],
    db_mem_size: str,
    db_disk_size: str,
    nameserver: tuple[str, ...],
    docker_runtime: Optional[str],
    docker_db_image_version: str,
    docker_db_image_name: str,
    db_os_access: Optional[str],
    create_certificates: bool,
    additional_db_parameter: tuple[str, ...],
    docker_environment_variable: tuple[str, ...],
    accelerator: tuple[str, ...],
    source_docker_repository_name: str,
    source_docker_tag_prefix: str,
    source_docker_username: Optional[str],
    source_docker_password: Optional[str],
    target_docker_repository_name: str,
    target_docker_tag_prefix: str,
    target_docker_username: Optional[str],
    target_docker_password: Optional[str],
    output_directory: str,
    temporary_base_directory: str,
    workers: int,
    task_dependencies_dot_file: Optional[str],
    log_level: Optional[str],
    use_job_specific_log_file: bool,
    bucketfs_http_port_forward: Optional[int],
    bucketfs_https_port_forward: Optional[int],
):
    """
    This command spawns a test environment with a docker-db container and a connected test-container.
    The test-container is reachable by the database for output redirects of UDFs.
    """
    with TerminationHandler():
        try:
            api.spawn_test_environment(
                environment_name,
                database_port_forward,
                bucketfs_port_forward,
                ssh_port_forward,
                db_mem_size,
                db_disk_size,
                nameserver,
                docker_runtime,
                docker_db_image_version,
                docker_db_image_name,
                db_os_access,
                create_certificates,
                additional_db_parameter,
                docker_environment_variable,
                accelerator,
                source_docker_repository_name,
                source_docker_tag_prefix,
                source_docker_username,
                source_docker_password,
                target_docker_repository_name,
                target_docker_tag_prefix,
                target_docker_username,
                target_docker_password,
                output_directory,
                temporary_base_directory,
                workers,
                task_dependencies_dot_file,
                log_level=log_level,
                use_job_specific_log_file=use_job_specific_log_file,
                bucketfs_http_port_forward=bucketfs_http_port_forward,
                bucketfs_https_port_forward=bucketfs_https_port_forward,
            )
        except ArgumentConstraintError as e:
            handle_wrong_argument_error(*e.args)


def handle_wrong_argument_error(argument_name, message):
    formatted = f"--{argument_name}".replace("_", "-")
    print(f"{formatted}: {message}")
    ctx = click.get_current_context()
    click.echo(ctx.get_help())
    exit(1)



from pathlib import Path
from typing import (
    Any,
)

import luigi

from exasol_integration_test_docker_environment.lib.base.base_task import BaseTaskType
from exasol_integration_test_docker_environment.lib.base.dependency_logger_base_task import (
    DependencyLoggerBaseTask,
)
from exasol_integration_test_docker_environment.lib.base.docker_base_task import (
    DockerBaseTask,
)


class FlavorsBaseTask(DependencyLoggerBaseTask):
    flavor_paths: tuple[str] = luigi.ListParameter()

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        for flavor_path in self.flavor_paths:
            if not Path(flavor_path).is_dir():
                raise OSError("Flavor path %s not a directory." % flavor_path)

    def create_tasks_for_flavors_with_common_params(
        self, cls, **kwargs
    ) -> dict[str, Any]:
        return {
            flavor_path: self._create_task_with_common_params(cls, flavor_path, kwargs)
            for flavor_path in self.flavor_paths
        }

    def _create_task_with_common_params(
        self, cls: type[BaseTaskType], flavor_path, kwargs
    ) -> BaseTaskType:
        params = {**kwargs, "flavor_path": flavor_path}
        task = self.create_child_task_with_common_params(cls, **params)
        return task


class FlavorBaseTask(DockerBaseTask):
    flavor_path: str = luigi.Parameter()

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        if not Path(self.flavor_path).is_dir():
            raise OSError("Flavor path %s not a directory." % self.flavor_path)

    def get_flavor_name(self) -> str:
        path = Path(self.flavor_path)
        flavor_name = path.name
        return flavor_name



from pathlib import Path

from exasol_integration_test_docker_environment.lib.base.task_logger_wrapper import (
    TaskLoggerWrapper,
)
from exasol_integration_test_docker_environment.lib.models.config.log_config import (
    log_config,
)


class AbstractLogHandler:

    def __init__(self, log_file_path: Path, logger: TaskLoggerWrapper) -> None:
        self._log_file_path = log_file_path
        self._logger = logger
        self._complete_log: list[str] = []
        self._error_message = None
        self._log_config = log_config()

    def __enter__(self):
        if self._log_file_path is not None:
            self._log_file = self._log_file_path.open("w")
        return self

    def __exit__(self, type, value, traceback) -> None:
        if self._log_file_path is not None:
            self._log_file.close()
        self.finish()

    def handle_log_lines(self, log_lines, error: bool = False) -> None:
        log_lines = log_lines.decode("utf-8")
        log_lines = log_lines.strip("\r\n")
        for log_line in log_lines.split("\n"):
            log_line = log_line.strip("\r\n")
            self.handle_log_line(log_line, error)

    def handle_log_line(self, log_line, error: bool = False) -> None:
        pass

    def finish(self) -> None:
        pass



import logging
from collections.abc import Iterable

from exasol_integration_test_docker_environment.lib.docker import ContextDockerClient


def remove_docker_networks(networks: Iterable[str]):
    """
    Removes the given networks using docker API.
    """
    with ContextDockerClient() as docker_client:
        for network in networks:
            try:
                docker_client.networks.get(network).remove()
            except Exception as e:
                logging.error(e)



import jsonpickle
from luigi.parameter import (
    Parameter,
    ParameterVisibility,
    _no_value,
)


class JsonPickleParameter(Parameter):

    def __init__(
        self,
        cls,
        default=_no_value,
        is_global=False,
        significant=True,
        description=None,
        config_path=None,
        positional=True,
        always_in_help=False,
        batch_method=None,
        is_optional=False,
        visibility=ParameterVisibility.PUBLIC,
    ) -> None:
        super().__init__(
            default,
            is_global,
            significant,
            description,
            config_path,
            positional,
            always_in_help,
            batch_method,
            visibility,
        )
        self.cls = cls
        self._is_optional = is_optional

    def parse(self, s) -> None:
        jsonpickle.set_preferred_backend("simplejson")
        loaded_object = jsonpickle.decode(s)
        if self._is_optional and loaded_object is None:
            return None
        elif not isinstance(loaded_object, self.cls):
            raise TypeError(
                "Type {} of loaded object does not match {}".format(
                    type(loaded_object), self.cls
                )
            )
        return loaded_object

    def serialize(self, x):
        jsonpickle.set_preferred_backend("simplejson")
        return jsonpickle.encode(x)



from exasol_integration_test_docker_environment.lib.base.task_state import TaskState


class WrongTaskStateException(Exception):

    def __init__(self, task_state: TaskState, method: str) -> None:
        super().__init__(
            f"Calling method {method} in task state {task_state} not allowed"
        )



import math
from pathlib import Path
from typing import (
    Optional,
    Union,
)

import docker
import humanfriendly
import importlib_resources
import luigi
import netaddr
from docker.client import DockerClient
from docker.models.containers import Container
from docker.models.volumes import Volume
from docker.types import DeviceRequest
from importlib_resources.abc import Traversable
from jinja2 import Template

from exasol_integration_test_docker_environment.lib import PACKAGE_NAME
from exasol_integration_test_docker_environment.lib.base.docker_base_task import (
    DockerBaseTask,
)
from exasol_integration_test_docker_environment.lib.base.json_pickle_parameter import (
    JsonPickleParameter,
)
from exasol_integration_test_docker_environment.lib.base.ssh_access import (
    SshKey,
    SshKeyCache,
)
from exasol_integration_test_docker_environment.lib.base.still_running_logger import (
    StillRunningLogger,
)
from exasol_integration_test_docker_environment.lib.docker.images.create.utils.pull_log_handler import (
    PullLogHandler,
)
from exasol_integration_test_docker_environment.lib.docker.images.image_info import (
    ImageInfo,
)
from exasol_integration_test_docker_environment.lib.models.data.container_info import (
    ContainerInfo,
)
from exasol_integration_test_docker_environment.lib.models.data.database_info import (
    DatabaseInfo,
)
from exasol_integration_test_docker_environment.lib.models.data.docker_network_info import (
    DockerNetworkInfo,
)
from exasol_integration_test_docker_environment.lib.models.data.ssh_info import SshInfo
from exasol_integration_test_docker_environment.lib.test_environment.db_version import (
    DbVersion,
)
from exasol_integration_test_docker_environment.lib.test_environment.docker_container_copy import (
    DockerContainerCopy,
)
from exasol_integration_test_docker_environment.lib.test_environment.parameter.docker_db_test_environment_parameter import (
    DbOsAccess,
    DockerDBTestEnvironmentParameter,
)
from exasol_integration_test_docker_environment.lib.test_environment.ports import (
    Ports,
    find_free_ports,
)

CERTIFICATES_MOUNT_DIR = "/certificates"
CERTIFICATES_DEFAULT_DIR = "/exa/etc/ssl/"


def int_or_none(value: Optional[str]) -> Optional[int]:
    return None if value is None else int(value)


class SpawnTestDockerDatabase(DockerBaseTask, DockerDBTestEnvironmentParameter):
    environment_name: str = luigi.Parameter()
    db_container_name: str = luigi.Parameter()
    attempt: int = luigi.IntParameter(default=1)
    network_info: DockerNetworkInfo = JsonPickleParameter(DockerNetworkInfo, significant=False)  # type: ignore
    ip_address_index_in_subnet: int = luigi.IntParameter(significant=False)
    docker_runtime: Optional[str] = luigi.OptionalParameter(
        default=None, significant=False
    )
    certificate_volume_name: Optional[str] = luigi.OptionalParameter(
        default=None, significant=False
    )
    additional_db_parameter: tuple[str, ...] = luigi.ListParameter()
    docker_environment_variables: tuple[str, ...] = luigi.ListParameter()
    accelerator: tuple[str, ...] = luigi.ListParameter()
    ssh_user: str = luigi.Parameter(default="root")
    ssh_key_file: Union[str, Path, None] = luigi.OptionalParameter(
        default=None, significant=False
    )

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        if self.ip_address_index_in_subnet < 0:
            raise Exception(
                "ip_address_index_in_subnet needs to be greater than 0 got %s"
                % self.ip_address_index_in_subnet
            )

        self.db_version = DbVersion.from_db_version_str(self.docker_db_image_version)
        self.docker_db_config_resource_name = f"docker_db_config/{self.db_version}"
        self.internal_ports = Ports.default_ports
        if self.ssh_port_forward is None:
            self.ssh_port_forward = str(find_free_ports(1)[0])

        def first_of(*args):
            return next(x for x in args if x is not None)

        bucketfs_http_port: int = int(
            first_of(
                self.bucketfs_http_port_forward,
                self.bucketfs_port_forward,
                Ports.default_ports.bucketfs_http,
            )
        )
        bucketfs_https_port: int = int(
            first_of(
                self.bucketfs_https_port_forward, Ports.default_ports.bucketfs_https
            )
        )

        self.forwarded_ports = Ports(
            database=int_or_none(self.database_port_forward),
            bucketfs=bucketfs_http_port,
            ssh=int_or_none(self.ssh_port_forward),
            bucketfs_https=bucketfs_https_port,
        )

    def run_task(self) -> None:
        subnet = netaddr.IPNetwork(self.network_info.subnet)
        db_ip_address = str(subnet[2 + self.ip_address_index_in_subnet])
        db_private_network = f"{db_ip_address}/{subnet.prefixlen}"
        database_info = None
        if self.network_info.reused:
            database_info = self._try_to_reuse_database(db_ip_address)
        if database_info is None:
            database_info = self._create_database_container(
                db_ip_address, db_private_network
            )
        self.return_object(database_info)

    def _try_to_reuse_database(self, db_ip_address: str) -> Optional[DatabaseInfo]:
        self.logger.info("Try to reuse database container %s", self.db_container_name)
        try:
            database_info = self._create_database_info(
                db_ip_address=db_ip_address, reused=True
            )
            return database_info
        except Exception as e:
            self.logger.warning(
                "Tried to reuse database container %s, but got Exeception %s. "
                "Fallback to create new database.",
                self.db_container_name,
                e,
            )
        return None

    def _get_ssh_key(self) -> SshKey:
        if self.ssh_key_file:
            return SshKey.read_from(self.ssh_key_file)
        self.ssh_key_file = SshKeyCache().private_key
        return SshKey.from_cache()

    def _handle_output(self, output_generator, image_info: ImageInfo) -> None:
        log_file_path = self.get_log_path().joinpath("pull_docker_db_image.log")
        with PullLogHandler(log_file_path, self.logger, image_info) as log_handler:
            still_running_logger = StillRunningLogger(
                self.logger, "pull image %s" % image_info.get_source_complete_name()
            )
            for log_line in output_generator:
                still_running_logger.log()
                log_handler.handle_log_lines(log_line)

    def _get_network_aliases(self) -> list[str]:
        network_aliases = [
            "exasol_test_database",
            "exasol-test-database",
            self.db_container_name,
        ]
        return network_aliases

    def _connect_docker_network(
        self,
        docker_client: DockerClient,
        container: Container,
        ip_address: str,
    ) -> None:
        network = docker_client.networks.get(self.network_info.network_name)
        aliases = self._get_network_aliases()
        network.connect(container, ipv4_address=ip_address, aliases=aliases)

    def _port_mapping(
        self, internal_ports: Ports, forwarded_ports: Ports
    ) -> dict[str, tuple[str, str]]:
        result = {}
        for name, internal in internal_ports.__dict__.items():
            forward = forwarded_ports.__getattribute__(name)
            if forward:
                result[f"{internal}/tcp"] = ("0.0.0.0", forward)
        return result

    def _create_database_container(
        self, db_ip_address: str, db_private_network: str
    ) -> DatabaseInfo:
        def get_authorized_keys(ssh_key) -> str:
            """
            Multiple authorized_keys can be comma-separated.
            """
            if self.db_os_access != DbOsAccess.SSH:
                return ""
            return ssh_key.public_key_as_string("itde-ssh-access")

        def enable_ssh_access(container: Container, authorized_keys: str):
            copy = DockerContainerCopy(container)
            copy.add_string_to_file(".ssh/authorized_keys", authorized_keys)
            copy.copy("/root/")

        self.logger.info("Starting database container %s", self.db_container_name)
        ssh_key = self._get_ssh_key()
        authorized_keys = get_authorized_keys(ssh_key)
        with self._get_docker_client() as docker_client:
            try:
                docker_client.containers.get(self.db_container_name).remove(
                    force=True, v=True
                )
            except:
                pass
            docker_db_image_info = self._pull_docker_db_images_if_necessary()
            db_volume = self._prepare_db_volume(
                docker_client,
                db_private_network,
                authorized_keys,
                docker_db_image_info,
            )
            port_mapping = self._port_mapping(self.internal_ports, self.forwarded_ports)
            volumes = {db_volume.name: {"bind": "/exa", "mode": "rw"}}
            if self.certificate_volume_name is not None:
                volumes[self.certificate_volume_name] = {
                    "bind": CERTIFICATES_MOUNT_DIR,
                    "mode": "rw",
                }
            device_requests = self._build_device_requests()
            db_container = docker_client.containers.create(
                image="%s" % (docker_db_image_info.get_source_complete_name()),
                name=self.db_container_name,
                detach=True,
                privileged=True,
                volumes=volumes,
                network_mode=None,
                ports=port_mapping,
                runtime=self.docker_runtime,
                environment=list(self.docker_environment_variables),
                device_requests=device_requests,
            )
            enable_ssh_access(db_container, authorized_keys)
            self._connect_docker_network(docker_client, db_container, db_ip_address)
            db_container.start()
            database_info = self._create_database_info(
                db_ip_address=db_ip_address, reused=False
            )
            return database_info

    def _create_database_info(self, db_ip_address: str, reused: bool) -> DatabaseInfo:
        with self._get_docker_client() as docker_client:
            db_container = docker_client.containers.get(self.db_container_name)
            if db_container.status != "running":
                raise Exception(f"Container {self.db_container_name} not running")
            network_aliases = self._get_network_aliases()
            container_info = ContainerInfo(
                container_name=self.db_container_name,
                ip_address=db_ip_address,
                network_aliases=network_aliases,
                network_info=self.network_info,
                volume_name=self._get_db_volume_name(),
            )
            ssh_info = SshInfo(self.ssh_user, str(self.ssh_key_file or ""))
            database_info = DatabaseInfo(
                host=db_ip_address,
                ports=self.internal_ports,
                reused=reused,
                container_info=container_info,
                ssh_info=ssh_info,
                forwarded_ports=self.forwarded_ports,
            )
            return database_info

    def _pull_docker_db_images_if_necessary(self) -> ImageInfo:
        image_name = "exasol/docker-db"
        assert self.docker_db_image_version is not None
        docker_db_image_info = ImageInfo(
            target_repository_name=image_name,
            source_repository_name=image_name,
            source_tag=self.docker_db_image_version,
            target_tag=self.docker_db_image_version,
            hash_value="",
            commit="",
            image_description=None,
        )
        with self._get_docker_client() as docker_client:
            try:
                docker_client.images.get(
                    docker_db_image_info.get_source_complete_name()
                )
            except docker.errors.ImageNotFound as e:
                self.logger.info(
                    "Pulling docker-db image %s",
                    docker_db_image_info.get_source_complete_name(),
                )
                output_generator = docker_client.api.pull(
                    docker_db_image_info.source_repository_name,
                    tag=docker_db_image_info.source_tag,
                    stream=True,
                )
                self._handle_output(output_generator, docker_db_image_info)
        return docker_db_image_info

    def _prepare_db_volume(
        self,
        docker_client,
        db_private_network: str,
        authorized_keys: str,
        docker_db_image_info: ImageInfo,
    ) -> Volume:
        volume, container = self._prepare_volume(
            docker_client,
            self._get_db_volume_name(),
            self._get_db_volume_preparation_container_name(),
            remove_old_instances=True,
        )
        try:
            self._upload_init_db_files(container, db_private_network, authorized_keys)
            self._execute_init_db(volume, container)
            return volume
        finally:
            container.remove(force=True)

    def _get_db_volume_preparation_container_name(self):
        return f"""{self.db_container_name}_preparation"""

    def _get_db_volume_name(self):
        return f"""{self.db_container_name}_volume"""

    def _remove_container(self, container_name: str) -> None:
        try:
            with self._get_docker_client() as docker_client:
                docker_client.containers.get(container_name).remove(force=True)
                self.logger.info("Removed container %s", container_name)
        except docker.errors.NotFound:
            pass

    def _remove_volume(self, volume_name: str) -> None:
        try:
            with self._get_docker_client() as docker_client:
                docker_client.volumes.get(volume_name).remove(force=True)
                self.logger.info("Removed volume %s", volume_name)
        except docker.errors.NotFound:
            pass

    def _prepare_volume(
        self,
        docker_client: docker.api.APIClient,
        volume_name,
        container_name,
        remove_old_instances: bool = False,
    ) -> tuple[Volume, Container]:
        """
        Create an intermediate Docker Container containing a volume that
        can be mounted into another Docker Container.
        """
        if remove_old_instances:
            self._remove_container(container_name)
            self._remove_volume(volume_name)
        volume = docker_client.volumes.create(volume_name)
        container = docker_client.containers.run(
            image="ubuntu:22.04",
            name=container_name,
            auto_remove=True,
            command="sleep infinity",
            detach=True,
            volumes={volume.name: {"bind": "/exa", "mode": "rw"}},
            labels={
                "test_environment_name": self.environment_name,
                "container_type": "db_volume_preparation_container",
            },
        )
        return volume, container

    def _db_file(self, filename: str) -> Traversable:
        return (
            importlib_resources.files(PACKAGE_NAME)
            / self.docker_db_config_resource_name
            / filename
        )

    def _upload_init_db_files(
        self,
        container: Container,
        db_private_network: str,
        authorized_keys: str,
    ) -> None:
        copy = DockerContainerCopy(container)
        init_script = self._db_file("init_db.sh")
        copy.add_string_to_file("init_db.sh", init_script.read_text())
        self._add_exa_conf(copy, db_private_network, authorized_keys)
        copy.copy("/")

    def _add_exa_conf(
        self,
        copy: DockerContainerCopy,
        db_private_network: str,
        authorized_keys: str,
    ) -> None:
        """
        Multiple authorized_keys can be comma-separated.
        """
        certificate_dir = (
            CERTIFICATES_MOUNT_DIR
            if self.certificate_volume_name is not None
            else CERTIFICATES_DEFAULT_DIR
        )
        template_file = self._db_file("EXAConf")
        template = Template(template_file.read_text())
        additional_db_parameter_str = " ".join(self.additional_db_parameter)
        rendered_template = template.render(
            private_network=db_private_network,
            db_version=str(self.db_version),
            db_port=self.internal_ports.database,
            ssh_port=self.internal_ports.ssh,
            bucketfs_http_port=self.internal_ports.bucketfs_http,
            bucketfs_https_port=self.internal_ports.bucketfs_https,
            image_version=self.docker_db_image_version,
            mem_size=self.mem_size,
            disk_size=self.disk_size,
            name_servers=",".join(self.nameservers),
            certificate_dir=certificate_dir,
            additional_db_parameters=additional_db_parameter_str,
            authorized_keys=authorized_keys,
        )
        copy.add_string_to_file("EXAConf", rendered_template)

    def _execute_init_db(
        self, db_volume: Volume, preparation_container: Container
    ) -> None:
        disk_size_in_bytes = humanfriendly.parse_size(self.disk_size)
        min_overhead_in_gigabyte = 2  # Exasol needs at least a 2 GB larger device than the configured disk size
        overhead_factor = max(
            0.01, (min_overhead_in_gigabyte * 1024 * 1024 * 1024) / disk_size_in_bytes
        )  # and in general 1% larger
        device_size_in_bytes = disk_size_in_bytes * (1 + overhead_factor)
        device_size_in_megabytes = math.ceil(
            device_size_in_bytes / (1024 * 1024)
        )  # The init_db.sh script works with MB, because its faster
        self.logger.info(
            f"Creating database volume of size {device_size_in_megabytes / 1024} GB using and overhead factor of {overhead_factor}"
        )
        (exit_code, output) = preparation_container.exec_run(
            cmd=f"bash /init_db.sh {device_size_in_megabytes}"
        )
        if exit_code != 0:
            raise Exception(
                "Error during preparation of docker-db volume {} got following output {}".format(
                    db_volume.name, output
                )
            )

    def cleanup_task(self, success) -> None:
        if (success and not self.no_database_cleanup_after_success) or (
            not success and not self.no_database_cleanup_after_failure
        ):
            volume_container = self._get_db_volume_preparation_container_name()
            try:
                self.logger.info(f"Cleaning up container %s", volume_container)
                self._remove_container(volume_container)
            except Exception as e:
                self.logger.error(
                    f"Error during removing container %s: %s", volume_container, e
                )

            try:
                self.logger.info(f"Cleaning up container %s", self.db_container_name)
                self._remove_container(self.db_container_name)
            except Exception as e:
                self.logger.error(
                    f"Error during removing container %s: %s", self.db_container_name, e
                )

            db_volume_name = self._get_db_volume_name()
            try:
                self.logger.info(f"Cleaning up docker volume %s", db_volume_name)
                self._remove_volume(db_volume_name)
            except Exception as e:
                self.logger.error(
                    f"Error during removing docker volume %s: %s", db_volume_name, e
                )

    def _build_device_requests(self) -> list[DeviceRequest]:
        if "nvidia=all" in self.accelerator:
            return [
                DeviceRequest(
                    driver="nvidia",
                    count=-1,
                    capabilities=[["gpu"]],
                )
            ]
        else:
            return []



import os
import re
import unittest
from typing import (
    Callable,
    Optional,
)

from exasol_integration_test_docker_environment.lib.docker import ContextDockerClient

INTEGRATION_TEST_DOCKER_ENVIRONMENT_DEFAULT_BIN = "itde"


def close_environments(*args):
    for env in args:
        try:
            if env is not None:
                env.close()
        except Exception as e:
            print(e)


def check_db_version_from_env() -> Optional[str]:
    retval = None
    if "EXASOL_VERSION" in os.environ and os.environ["EXASOL_VERSION"] != "default":
        retval = os.environ["EXASOL_VERSION"]
    return retval


def multiassert(assert_list: list[Callable], unit_test: unittest.TestCase):
    failure_log: list[str] = []
    for assert_fn in assert_list:
        try:
            assert_fn()
        except AssertionError as e:
            failure_log.append(f"\nFailure {len(failure_log)}: {str(e)}")

    if len(failure_log) != 0:
        res_failure_log = "\n".join(failure_log)
        unit_test.fail(f"{len(failure_log)} failures within test.\n {res_failure_log}")


def find_docker_container_names(container_name_substr: str) -> list[str]:
    with ContextDockerClient() as docker_client:
        containers = [
            c.name
            for c in docker_client.containers.list()
            if container_name_substr in c.name
        ]
        return containers


def normalize_request_name(name: str):
    name = name.split("/")[-1]
    name = re.sub(r"[\[\]._]+", "_", name)
    return name.strip("_")



#! /usr/bin/env python3
#
from exasol_integration_test_docker_environment.cli.cli import cli


def main():
    # required so the cli will print the available subcommands
    from exasol_integration_test_docker_environment.cli.commands import (
        environment,
        health,
        spawn_test_environment,
    )

    cli()


if __name__ == "__main__":
    main()
